# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class RandomRegular(Graph):
    r"""Random k-regular graph.

    The random regular graph has the property that every node is connected to
    k other nodes. That graph is simple (without loops or double edges),
    k-regular (each vertex is adjacent to k nodes), and undirected.

    Parameters
    ----------
    N : int
        Number of nodes (default is 64)
    k : int
        Number of connections, or degree, of each node (default is 6)
    max_iter : int
        Maximum number of iterations (default is 10)
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Notes
    -----
    The *pairing model* algorithm works as follows. First create n*d *half
    edges*. Then repeat as long as possible: pick a pair of half edges and if
    it's legal (doesn't create a loop nor a double edge) add it to the graph.

    References
    ----------
    See :cite:`kim2003randomregulargraphs`.
    This code has been adapted from matlab to python.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.RandomRegular(N=64, k=5, seed=42)
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=64, k=6, max_iter=10, seed=None, **kwargs):

        self.k = k
        self.max_iter = max_iter
        self.seed = seed

        self.logger = utils.build_logger(__name__)

        rng = np.random.default_rng(seed)

        # continue until a proper graph is formed
        if (N * k) % 2 == 1:
            raise ValueError("input error: N*d must be even!")

        # a list of open half-edges
        U = np.kron(np.ones(k), np.arange(N))

        # the graphs adjacency matrix
        A = sparse.lil_matrix(np.zeros((N, N)))

        edgesTested = 0
        repetition = 1

        while np.size(U) and repetition < max_iter:
            edgesTested += 1

            if edgesTested % 5000 == 0:
                self.logger.debug("createRandRegGraph() progress: edges= "
                                  "{}/{}.".format(edgesTested, N*k/2))

            # chose at random 2 half edges
            i1 = rng.integers(0, U.shape[0])
            i2 = rng.integers(0, U.shape[0])
            v1 = U[i1]
            v2 = U[i2]

            # check that there are no loops nor parallel edges
            if v1 == v2 or A[v1, v2] == 1:
                # restart process if needed
                if edgesTested == N*k:
                    repetition = repetition + 1
                    edgesTested = 0
                    U = np.kron(np.ones(k), np.arange(N))
                    A = sparse.lil_matrix(np.zeros((N, N)))
            else:
                # add edge to graph
                A[v1, v2] = 1
                A[v2, v1] = 1

                # remove used half-edges
                v = sorted([i1, i2])
                U = np.concatenate((U[:v[0]], U[v[0] + 1:v[1]], U[v[1] + 1:]))

        super(RandomRegular, self).__init__(A, **kwargs)

        self.is_regular()

    def is_regular(self):
        r"""
        Troubleshoot a given regular graph.

        """
        warn = False
        msg = 'The given matrix'

        # check symmetry
        if np.abs(self.A - self.A.T).sum() > 0:
            warn = True
            msg = '{} is not symmetric,'.format(msg)

        # check parallel edged
        if self.A.max(axis=None) > 1:
            warn = True
            msg = '{} has parallel edges,'.format(msg)

        # check that d is d-regular
        if np.min(self.d) != np.max(self.d):
            warn = True
            msg = '{} is not d-regular,'.format(msg)

        # check that g doesn't contain any self-loop
        if self.A.diagonal().any():
            warn = True
            msg = '{} has self loop.'.format(msg)

        if warn:
            self.logger.warning('{}.'.format(msg[:-1]))

    def _get_extra_repr(self):
        return dict(k=self.k, seed=self.seed)



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class SwissRoll(Graph):
    r"""Sampled Swiss roll manifold.

    Parameters
    ----------
    N : int
        Number of vertices (default = 400)
    a : int
        (default = 1)
    b : int
        (default = 4)
    dim : int
        (default = 3)
    thresh : float
        (default = 1e-6)
    s : float
        sigma (default =  sqrt(2./N))
    noise : bool
        Wether to add noise or not (default = False)
    srtype : str
        Swiss roll Type, possible arguments are 'uniform' or 'classic'
        (default = 'uniform')
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.SwissRoll(N=200, seed=42)
    >>> fig = plt.figure()
    >>> ax1 = fig.add_subplot(121)
    >>> ax2 = fig.add_subplot(122, projection='3d')
    >>> _ = ax1.spy(G.W, markersize=1)
    >>> _ = G.plot(ax=ax2)

    """

    def __init__(self, N=400, a=1, b=4, dim=3, thresh=1e-6, s=None,
                 noise=False, srtype='uniform', seed=None, **kwargs):

        if s is None:
            s = np.sqrt(2. / N)

        self.a = a
        self.b = b
        self.dim = dim
        self.thresh = thresh
        self.s = s
        self.noise = noise
        self.srtype = srtype
        self.seed = seed

        rng = np.random.default_rng(seed)
        y1 = rng.uniform(size=N)
        y2 = rng.uniform(size=N)

        if srtype == 'uniform':
            tt = np.sqrt((b * b - a * a) * y1 + a * a)
        elif srtype == 'classic':
            tt = (b - a) * y1 + a
        tt *= np.pi

        if dim == 2:
            x = np.array((tt * np.cos(tt), tt * np.sin(tt)))
        elif dim == 3:
            x = np.array((tt * np.cos(tt), 21 * y2, tt * np.sin(tt)))

        if noise:
            x += rng.normal(size=x.shape)

        self.x = x
        self.dim = dim

        coords = utils.rescale_center(x)
        dist = utils.distanz(coords)
        W = np.exp(-np.power(dist, 2) / (2. * s**2))
        W -= np.diag(np.diag(W))
        W[W < thresh] = 0

        plotting = {
            'vertex_size': 60,
            'limits': np.array([-1, 1, -1, 1, -1, 1]),
            'elevation': 15,
            'azimuth': -90,
            'distance': 7,
        }

        super(SwissRoll, self).__init__(W, coords=coords.T,
                                        plotting=plotting,
                                        **kwargs)

    def _get_extra_repr(self):
        return {'a': self.a,
                'b': self.b,
                'dim': self.dim,
                'thresh': '{:.0e}'.format(self.thresh),
                's': '{:.2f}'.format(self.s),
                'noise': self.noise,
                'srtype': self.srtype,
                'seed': self.seed}



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class Minnesota(Graph):
    r"""Minnesota road network (from MatlabBGL).

    Parameters
    ----------
    connected : bool
        If True, the adjacency matrix is adjusted so that all edge weights are
        equal to 1, and the graph is connected. Set to False to get the
        original disconnected graph.

    References
    ----------
    See :cite:`gleich`.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Minnesota()
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.5)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, connected=True, **kwargs):

        self.connected = connected

        data = utils.loadmat('pointclouds/minnesota')
        self.labels = data['labels']
        A = data['A']

        plotting = {"limits": np.array([-98, -89, 43, 50]),
                    "vertex_size": 40}

        if connected:

            # Missing edges needed to connect the graph.
            A = sparse.lil_matrix(A)
            A[348, 354] = 1
            A[354, 348] = 1
            A = sparse.csc_matrix(A)

            # Binarize: 8 entries are equal to 2 instead of 1.
            A = (A > 0).astype(bool)

        super(Minnesota, self).__init__(A, coords=data['xy'],
                                        plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        return dict(connected=self.connected)



r"""
Concentration of the eigenvalues
================================

The eigenvalues of the graph Laplacian concentrates to the same value as the
graph becomes full.
"""

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

n_neighbors = [1, 2, 5, 8]
fig, axes = plt.subplots(3, len(n_neighbors), figsize=(15, 8))


for k, ax in zip(n_neighbors, axes.T):

    graph = pg.graphs.Ring(17, k=k)
    graph.compute_fourier_basis()
    graph.plot(graph.U[:, 1], ax=ax[0])
    ax[0].axis('equal')
    ax[1].spy(graph.W)
    ax[2].plot(graph.e, '.')
    ax[2].set_title('k={}'.format(k))
    #graph.set_coordinates('line1D')
    #graph.plot(graph.U[:, :4], ax=ax[3], title='')

    # Check that the DFT matrix is an eigenbasis of the Laplacian.
    U = np.fft.fft(np.identity(graph.n_vertices))
    LambdaM = (graph.L.todense().dot(U)) / (U + 1e-15)
    # Eigenvalues should be real.
    assert np.all(np.abs(np.imag(LambdaM)) < 1e-10)
    LambdaM = np.real(LambdaM)
    # Check that the eigenvectors are really eigenvectors of the laplacian.
    Lambda = np.mean(LambdaM, axis=0)
    assert np.all(np.abs(LambdaM - Lambda) < 1e-10)

fig.tight_layout()
plt.show()



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class Logo(Graph):
    r"""GSP logo.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Logo()
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.5)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, **kwargs):

        data = utils.loadmat('pointclouds/logogsp')

        # Remove 1 because the index in python start at 0 and not at 1
        self.info = {"idx_g": data["idx_g"]-1,
                     "idx_s": data["idx_s"]-1,
                     "idx_p": data["idx_p"]-1}

        plotting = {"limits": np.array([0, 640, -400, 0])}

        super(Logo, self).__init__(data['W'], coords=data['coords'],
                                   plotting=plotting, **kwargs)



# -*- coding: utf-8 -*-

import traceback

import numpy as np
from scipy import sparse, spatial

from pygsp2 import utils
from pygsp2.graphs import Graph  # prevent circular import in Python < 3.5

_logger = utils.build_logger(__name__)


def _import_pfl():
    try:
        import pyflann as pfl
    except Exception as e:
        raise ImportError('Cannot import pyflann. Choose another nearest '
                          'neighbors method or try to install it with '
                          'pip (or conda) install pyflann (or pyflann3). '
                          'Original exception: {}'.format(e))
    return pfl


class NNGraph(Graph):
    r"""Nearest-neighbor graph from given point cloud.

    Parameters
    ----------
    Xin : ndarray
        Input points, Should be an `N`-by-`d` matrix, where `N` is the number
        of nodes in the graph and `d` is the dimension of the feature space.
    NNtype : string, optional
        Type of nearest neighbor graph to create. The options are 'knn' for
        k-Nearest Neighbors or 'radius' for epsilon-Nearest Neighbors (default
        is 'knn').
    use_flann : bool, optional
        Use Fast Library for Approximate Nearest Neighbors (FLANN) or not.
        (default is False)
    center : bool, optional
        Center the data so that it has zero mean (default is True)
    rescale : bool, optional
        Rescale the data so that it lies in a l2-sphere (default is True)
    k : int, optional
        Number of neighbors for knn (default is 10)
    sigma : float, optional
        Width of the similarity kernel.
        By default, it is set to the average of the nearest neighbor distance.
    epsilon : float, optional
        Radius for the epsilon-neighborhood search (default is 0.01)
    plotting : dict, optional
        Dictionary of plotting parameters. See :obj:`pygsp2.plotting`.
        (default is {})
    symmetrize_type : string, optional
        Type of symmetrization to use for the adjacency matrix. See
        :func:`pygsp2.utils.symmetrization` for the options.
        (default is 'average')
    dist_type : string, optional
        Type of distance to compute. See
        :func:`pyflann.index.set_distance_type` for possible options.
        (default is 'euclidean')
    order : float, optional
        Only used if dist_type is 'minkowski'; represents the order of the
        Minkowski distance. (default is 0)

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> X = np.random.default_rng(42).uniform(size=(30, 2))
    >>> G = graphs.NNGraph(X)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=5)
    >>> _ = G.plot(ax=axes[1])

    Notes
    -----
    The similarity kernel used is the gaussian kernel [1]:

    .. math::

        W_{i,j} = 
        \begin{cases}
            exp(-\frac{[dist(i,j)]^2}{2\sigma^2}) & \text{if dist(i,j)} < \epsilon \\
            0 & \text{otherwise}.
        \end{cases}

    References
    ----------
    .. [1] Shuman, D. I., Narang, S. K., Frossard, P., Ortega, A.,
        & Vandergheynst, P. (2013). The emerging field of signal processing on
        graphs: Extending high-dimensional data analysis to networks and other
        irregular domains. IEEE signal processing magazine, 30(3), 83-98.
    """

    def __init__(self, Xin, NNtype='knn', use_flann=False, center=True,
                 rescale=True, k=10, sigma=None, epsilon=0.01,
                 plotting={}, symmetrize_type='average', dist_type='euclidean',
                 order=0, **kwargs):

        self.Xin = Xin
        self.NNtype = NNtype
        self.use_flann = use_flann
        self.center = center
        self.rescale = rescale
        self.k = k
        self.sigma = sigma
        self.epsilon = epsilon
        self.symmetrize_type = symmetrize_type
        self.dist_type = dist_type
        self.order = order

        N, d = np.shape(self.Xin)
        Xout = self.Xin

        if k >= N:
            raise ValueError('The number of neighbors (k={}) must be smaller '
                             'than the number of nodes ({}).'.format(k, N))

        if self.center:
            Xout = self.Xin - np.kron(np.ones((N, 1)),
                                      np.mean(self.Xin, axis=0))

        if self.rescale:
            bounding_radius = 0.5 * np.linalg.norm(np.amax(Xout, axis=0) -
                                                   np.amin(Xout, axis=0), 2)
            scale = np.power(N, 1. / float(min(d, 3))) / 10.
            Xout *= scale / bounding_radius

        # Translate distance type string to corresponding Minkowski order.
        dist_translation = {"euclidean": 2,
                            "manhattan": 1,
                            "max_dist": np.inf,
                            "minkowski": order
                            }

        if self.NNtype == 'knn':
            spi = np.zeros((N * k))
            spj = np.zeros((N * k))
            spv = np.zeros((N * k))

            if self.use_flann:
                pfl = _import_pfl()
                pfl.set_distance_type(dist_type, order=order)
                flann = pfl.FLANN()

                # Default FLANN parameters (I tried changing the algorithm and
                # testing performance on huge matrices, but the default one
                # seems to work best).
                NN, D = flann.nn(Xout, Xout, num_neighbors=(k + 1),
                                 algorithm='kdtree')

            else:
                kdt = spatial.KDTree(Xout)
                D, NN = kdt.query(Xout, k=(k + 1),
                                  p=dist_translation[dist_type])

            if self.sigma is None:
                self.sigma = np.mean(D[:, 1:])  # Discard distance to self.

            for i in range(N):
                spi[i * k:(i + 1) * k] = np.kron(np.ones((k)), i)
                spj[i * k:(i + 1) * k] = NN[i, 1:]
                spv[i * k:(i + 1) * k] = np.exp(-np.square(D[i, 1:]) /
                                                (2.*np.square(self.sigma)))

        elif self.NNtype == 'radius':

            kdt = spatial.KDTree(Xout)
            D, NN = kdt.query(Xout, k=len(Xout), distance_upper_bound=epsilon,
                              p=dist_translation[dist_type])
            # Added to keep legacy functionality
            D2 = []
            NN2 = []
            for d, nn in zip(D, NN):
                keep_idx = np.where(d != np.inf)
                NN2.append(nn[:len(keep_idx[0])])
                D2.append(d[keep_idx])
            NN = NN2
            D = D2
            del D2, NN2

            if self.sigma is None:
                # Discard distance to self.
                self.sigma = np.mean([np.mean(d[1:]) for d in D])
            count = 0
            for i in range(N):
                count = count + len(NN[i])

            spi = np.zeros((count))
            spj = np.zeros((count))
            spv = np.zeros((count))

            start = 0
            for i in range(N):
                leng = len(NN[i]) - 1
                spi[start:start + leng] = np.kron(np.ones((leng)), i)
                spj[start:start + leng] = NN[i][1:]
                spv[start:start + leng] = np.exp(-np.square(D[i, 1:]) /
                                                 (2.*np.square(self.sigma)))
                start = start + leng

        else:
            raise ValueError('Unknown NNtype {}'.format(self.NNtype))

        W = sparse.csc_matrix((spv, (spi, spj)), shape=(N, N))

        # Sanity check
        if np.shape(W)[0] != np.shape(W)[1]:
            raise ValueError('Weight matrix W is not square')

        # Enforce symmetry. Note that checking symmetry with
        # np.abs(W - W.T).sum() is as costly as the symmetrization itself.
        W = utils.symmetrize(W, method=symmetrize_type)

        super(NNGraph, self).__init__(W, plotting=plotting,
                                      coords=Xout, **kwargs)

    def _get_extra_repr(self):
        return {'NNtype': self.NNtype,
                'use_flann': self.use_flann,
                'center': self.center,
                'rescale': self.rescale,
                'k': self.k,
                'sigma': '{:.2f}'.format(self.sigma),
                'epsilon': '{:.2f}'.format(self.epsilon),
                'symmetrize_type': self.symmetrize_type,
                'dist_type': self.dist_type,
                'order': self.order}



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


logger = utils.build_logger(__name__)


class LineGraph(Graph):
    r"""Build the line graph of a graph.

    Each vertex of the line graph represents an edge in the original graph. Two
    vertices are connected if the edges they represent share a vertex in the
    original graph.

    Parameters
    ----------
    graph : :class:`Graph`

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> graph = graphs.Sensor(5, k=2, seed=10)
    >>> line_graph = graphs.LineGraph(graph)
    >>> fig, ax = plt.subplots()
    >>> fig, ax = graph.plot('blue', edge_color='blue', indices=True, ax=ax)
    >>> fig, ax = line_graph.plot('red', edge_color='red', indices=True, ax=ax)
    >>> _ = ax.set_title('graph and its line graph')

    """

    def __init__(self, graph, **kwargs):

        if graph.is_weighted():
            logger.warning('Your graph is weighted, and is considered '
                           'unweighted to build a binary line graph.')

        graph.compute_differential_operator()
        # incidence = np.abs(graph.D)  # weighted?
        incidence = (graph.D != 0)

        adjacency = incidence.T.dot(incidence).astype(int)
        adjacency -= sparse.identity(graph.n_edges, dtype=int)

        try:
            coords = incidence.T.dot(graph.coords) / 2
        except AttributeError:
            coords = None

        super(LineGraph, self).__init__(adjacency, coords=coords,
                plotting=graph.plotting, **kwargs)



r"""
Metro Graph Signal
==================

Example of a graph signal for the Santiago Metro.

This example shows a graph signal defined over a graph induced by the Santiago
Metro (https://en.wikipedia.org/wiki/Santiago_Metro). Each station is a
node. Two nodes are connected if the corresponding stations are connected. The
data for each node is the daily average of people leaving each station.

To run this example, you need to download three files and place them in the
same directory as this script.

1. Download the file `Tablas de subidas y bajadas nov23.zip` from this link:

https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip

Then, uncompress the zip file and copy `2023.11 Matriz_baj_SS_MH.xlsb` to the
same location as this script.

2. Download the file `santiago_metro_stations_coords.geojson` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson

3. Download the file `santiago_metro_stations_connections.txt` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt
"""
# %%
import os
import sys
import pandas as pd
from unidecode import unidecode
import matplotlib.pyplot as plt

from pygsp2.utils_examples import (
    make_metro_graph, 
    metro_database_preprocessing, 
    plot_signal_in_graph,
    fetch_data
)

current_dir = os.getcwd()
os.chdir(current_dir)

# %% Load data
assets_dir = os.path.join(current_dir, 'data')
fetch_data(assets_dir, "metro")

try:
    commutes = pd.read_excel(
        os.path.join(assets_dir, '2023.11 Matriz_baj_SS_MH.xlsb'),
        header=1,
        sheet_name='bajadas_prom_laboral')
except FileNotFoundError:
    print(f'Data file was not found in:\n {os.getcwd()}')
    print('Download it from:\n' +
          r'https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip')
    sys.exit(1)

# %% Load graph

G, pos = make_metro_graph(
    edgesfile=os.path.join(assets_dir, 'santiago_metro_stations_connections.txt'),
    coordsfile=os.path.join(assets_dir, 'santiago_metro_stations_coords.geojson')
)

stations = [name.upper() for name in list(G.nodes)]
stations = [unidecode(station) for station in stations]

metro_commutes, signal = metro_database_preprocessing(commutes, stations)

# %% Let me know if there are stations not in the database

available_stations = list(metro_commutes['paradero'])

stations_missing = []
for i, station in enumerate(stations):
    if not station in available_stations:
        stations_missing.append(station)

print(f'{len(stations_missing)} Stations Not Meassured:')
_ = [print(f'\t{i+1}. {station}')
     for i, station in enumerate(stations_missing)]

# %% Plot signal in graph
fig, ax = plot_signal_in_graph(G, signal, 'Promedio Diario\nBajadas de Metro')
#fig.savefig('metro_graph_signal.png')
plt.show()


# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.reduction` module implements functionalities for the reduction
of graphs' vertex set while keeping the graph structure.

.. autosummary::

    tree_multiresolution
    graph_multiresolution
    kron_reduction
    pyramid_analysis
    pyramid_synthesis
    interpolate
    graph_sparsify

"""

import numpy as np
from scipy import sparse, stats
from scipy.sparse import linalg

from pygsp2 import graphs, filters, utils


logger = utils.build_logger(__name__)


def _analysis(g, s, **kwargs):
    # TODO: that is the legacy analysis method.
    s = g.filter(s, **kwargs)
    while s.ndim < 3:
        s = np.expand_dims(s, 1)
    return s.swapaxes(1, 2).reshape(-1, s.shape[1], order='F')


def graph_sparsify(M, epsilon, maxiter=10, seed=None):
    r"""Sparsify a graph (with Spielman-Srivastava).

    Parameters
    ----------
    M : Graph or sparse matrix
        Graph structure or a Laplacian matrix
    epsilon : float
        Sparsification parameter, which must be between ``1/sqrt(N)`` and 1.
    maxiter : int, optional
        Maximum number of iterations.
    seed : {None, int, RandomState, Generator}, optional
        Seed for the random number generator (for reproducible sparsification).

    Returns
    -------
    Mnew : Graph or sparse matrix
        New graph structure or sparse matrix

    Examples
    --------
    >>> from pygsp2 import reduction
    >>> from matplotlib import pyplot as plt
    >>> G = graphs.Sensor(100, k=20, distributed=True, seed=1)
    >>> Gs = reduction.graph_sparsify(G, epsilon=0.4, seed=1)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = G.plot(ax=axes[0], title='original')
    >>> Gs.coords = G.coords
    >>> _ = Gs.plot(ax=axes[1], title='sparsified')

    References
    ----------
    See :cite:`spielman2011graph`, :cite:`rudelson1999random` and :cite:`rudelson2007sampling`.
    for more informations

    """
    # Test the input parameters
    if isinstance(M, graphs.Graph):
        if not M.lap_type == 'combinatorial':
            raise NotImplementedError
        L = M.L
    else:
        L = M

    N = np.shape(L)[0]

    if not 1./np.sqrt(N) <= epsilon < 1:
        raise ValueError('GRAPH_SPARSIFY: Epsilon out of required range')

    # Not sparse
    resistance_distances = utils.resistance_distance(L).toarray()
    # Get the Weight matrix
    if isinstance(M, graphs.Graph):
        W = M.W
    else:
        W = np.diag(L.diagonal()) - L.toarray()
        W[W < 1e-10] = 0

    W = sparse.coo_matrix(W)
    W.data[W.data < 1e-10] = 0
    W = W.tocsc()
    W.eliminate_zeros()

    start_nodes, end_nodes, weights = sparse.find(sparse.tril(W))

    # Calculate the new weights.
    weights = np.maximum(0, weights)
    Re = np.maximum(0, resistance_distances[start_nodes, end_nodes])
    Pe = weights * Re
    Pe = Pe / np.sum(Pe)
    dist = stats.rv_discrete(values=(np.arange(len(Pe)), Pe), seed=seed)

    for i in range(maxiter):
        # Rudelson, 1996 Random Vectors in the Isotropic Position
        # (too hard to figure out actual C0)
        C0 = 1 / 30.
        # Rudelson and Vershynin, 2007, Thm. 3.1
        C = 4 * C0
        q = round(N * np.log(N) * 9 * C**2 / (epsilon**2))

        results = dist.rvs(size=int(q))
        spin_values, spin_counts = np.unique(results, return_counts=True)
        spin_counts = np.column_stack((spin_values, spin_counts)).astype(int)
        per_spin_weights = weights / (q * Pe)

        counts = np.zeros(np.shape(weights)[0])
        counts[spin_counts[:, 0]] = spin_counts[:, 1]
        new_weights = counts * per_spin_weights

        sparserW = sparse.csc_matrix((new_weights, (start_nodes, end_nodes)),
                                     shape=(N, N))
        sparserW = sparserW + sparserW.T
        sparserL = sparse.diags(sparserW.diagonal(), 0) - sparserW

        if graphs.Graph(sparserW).is_connected():
            break
        elif i == maxiter - 1:
            logger.warning('Despite attempts to reduce epsilon, sparsified graph is disconnected')
        else:
            epsilon -= (epsilon - 1/np.sqrt(N)) / 2.

    if isinstance(M, graphs.Graph):
        sparserW = sparse.diags(sparserL.diagonal(), 0) - sparserL
        if not M.is_directed():
            sparserW = (sparserW + sparserW.T) / 2.

        Mnew = graphs.Graph(sparserW)
        #M.copy_graph_attributes(Mnew)
    else:
        Mnew = sparse.lil_matrix(sparserL)

    return Mnew


def interpolate(G, f_subsampled, keep_inds, order=100, reg_eps=0.005, **kwargs):
    r"""Interpolate a graph signal.

    Parameters
    ----------
    G : Graph
    f_subsampled : ndarray
        A graph signal on the graph G.
    keep_inds : ndarray
        List of indices on which the signal is sampled.
    order : int
        Degree of the Chebyshev approximation (default = 100).
    reg_eps : float
        The regularized graph Laplacian is $\bar{L}=L+\epsilon I$.
        A smaller epsilon may lead to better regularization,
        but will also require a higher order Chebyshev approximation.

    Returns
    -------
    f_interpolated : ndarray
        Interpolated graph signal on the full vertex set of G.

    References
    ----------
    See :cite:`pesenson2009variational`

    """
    L_reg = G.L + reg_eps * sparse.eye(G.N)
    K_reg = getattr(G.mr, 'K_reg', kron_reduction(L_reg, keep_inds))
    green_kernel = getattr(G.mr, 'green_kernel',
                           filters.Filter(G, lambda x: 1. / (reg_eps + x)))

    alpha = K_reg.dot(f_subsampled)

    try:
        Nv = np.shape(f_subsampled)[1]
        f_interpolated = np.zeros((G.N, Nv))
    except IndexError:
        f_interpolated = np.zeros((G.N))

    f_interpolated[keep_inds] = alpha

    return _analysis(green_kernel, f_interpolated, order=order, **kwargs)


def graph_multiresolution(G, levels, sparsify=True, sparsify_eps=None,
                          downsampling_method='largest_eigenvector',
                          reduction_method='kron', compute_full_eigen=False,
                          reg_eps=0.005):
    r"""Compute a pyramid of graphs (by Kron reduction).

    'graph_multiresolution(G,levels)' computes a multiresolution of
    graph by repeatedly downsampling and performing graph reduction. The
    default downsampling method is the largest eigenvector method based on
    the polarity of the components of the eigenvector associated with the
    largest graph Laplacian eigenvalue. The default graph reduction method
    is Kron reduction followed by a graph sparsification step.
    *param* is a structure of optional parameters.

    Parameters
    ----------
    G : Graph structure
        The graph to reduce.
    levels : int
        Number of level of decomposition
    lambd : float
        Stability parameter. It adds self loop to the graph to give the
        algorithm some stability (default = 0.025). [UNUSED?!]
    sparsify : bool
        To perform a spectral sparsification step immediately after
        the graph reduction (default is True).
    sparsify_eps : float
        Parameter epsilon used in the spectral sparsification
        (default is min(10/sqrt(G.N),.3)).
    downsampling_method: string
        The graph downsampling method (default is 'largest_eigenvector').
    reduction_method : string
        The graph reduction method (default is 'kron')
    compute_full_eigen : bool
        To also compute the graph Laplacian eigenvalues and eigenvectors
        for every graph in the multiresolution sequence (default is False).
    reg_eps : float
        The regularized graph Laplacian is :math:`\bar{L}=L+\epsilon I`.
        A smaller epsilon may lead to better regularization, but will also
        require a higher order Chebyshev approximation. (default is 0.005)

    Returns
    -------
    Gs : list
        A list of graph layers.

    Examples
    --------
    >>> from pygsp2 import reduction
    >>> levels = 5
    >>> G = graphs.Sensor(N=512)
    >>> G.compute_fourier_basis()
    >>> Gs = reduction.graph_multiresolution(G, levels, sparsify=False)
    >>> for idx in range(levels):
    ...     fig, ax = Gs[idx].plot(title='Reduction level: {}'.format(idx))

    """
    if sparsify_eps is None:
        sparsify_eps = min(10. / np.sqrt(G.N), 0.3)

    if compute_full_eigen:
        G.compute_fourier_basis()
    else:
        G.estimate_lmax()

    Gs = [G]
    Gs[0].mr = {'idx': np.arange(G.N), 'orig_idx': np.arange(G.N)}

    for i in range(levels):
        if downsampling_method == 'largest_eigenvector':
            if Gs[i]._U is not None:
                V = Gs[i].U[:, -1]
            else:
                V = linalg.eigs(Gs[i].L, 1)[1][:, 0]

            V *= np.sign(V[0])
            ind = np.nonzero(V >= 0)[0]

        else:
            raise NotImplementedError('Unknown graph downsampling method.')

        if reduction_method == 'kron':
            Gs.append(kron_reduction(Gs[i], ind))

        else:
            raise NotImplementedError('Unknown graph reduction method.')

        if sparsify and Gs[i+1].N > 2:
            Gs[i+1] = graph_sparsify(Gs[i+1], min(max(sparsify_eps, 2. / np.sqrt(Gs[i+1].N)), 1.))
            # TODO : Make in place modifications instead!

        if compute_full_eigen:
            Gs[i+1].compute_fourier_basis()
        else:
            Gs[i+1].estimate_lmax()

        Gs[i+1].mr = {'idx': ind, 'orig_idx': Gs[i].mr['orig_idx'][ind], 'level': i}

        L_reg = Gs[i].L + reg_eps * sparse.eye(Gs[i].N)
        Gs[i].mr['K_reg'] = kron_reduction(L_reg, ind)
        Gs[i].mr['green_kernel'] = filters.Filter(Gs[i], lambda x: 1./(reg_eps + x))

    return Gs


def kron_reduction(G, ind):
    r"""Compute the Kron reduction.

    This function perform the Kron reduction of the weight matrix in the
    graph *G*, with boundary nodes labeled by *ind*. This function will
    create a new graph with a weight matrix Wnew that contain only boundary
    nodes and is computed as the Schur complement of the original matrix
    with respect to the selected indices.

    Parameters
    ----------
    G : Graph or sparse matrix
        Graph structure or weight matrix
    ind : list
        indices of the nodes to keep

    Returns
    -------
    Gnew : Graph or sparse matrix
        New graph structure or weight matrix


    References
    ----------
    See :cite:`dorfler2013kron`

    """
    if isinstance(G, graphs.Graph):

        if G.lap_type != 'combinatorial':
                msg = 'Unknown reduction for {} Laplacian.'.format(G.lap_type)
                raise NotImplementedError(msg)

        if G.is_directed():
            msg = 'This method only work for undirected graphs.'
            raise NotImplementedError(msg)

        L = G.L

    else:

        L = G

    N = np.shape(L)[0]
    ind_comp = np.setdiff1d(np.arange(N, dtype=int), ind)

    L_red = L[np.ix_(ind, ind)]
    L_in_out = L[np.ix_(ind, ind_comp)]
    L_out_in = L[np.ix_(ind_comp, ind)].tocsc()
    L_comp = L[np.ix_(ind_comp, ind_comp)].tocsc()

    Lnew = L_red - L_in_out.dot(linalg.spsolve(L_comp, L_out_in))

    # Make the laplacian symmetric if it is almost symmetric!
    if np.abs(Lnew - Lnew.T).sum() < np.spacing(1) * np.abs(Lnew).sum():
        Lnew = (Lnew + Lnew.T) / 2.

    if isinstance(G, graphs.Graph):
        # Suppress the diagonal ? This is a good question?
        Wnew = sparse.diags(Lnew.diagonal(), 0) - Lnew
        Snew = Lnew.diagonal() - np.ravel(Wnew.sum(0))
        if np.linalg.norm(Snew, 2) >= np.spacing(1000):
            Wnew = Wnew + sparse.diags(Snew, 0)

        # Removing diagonal for stability
        Wnew = Wnew - Wnew.diagonal()

        coords = G.coords[ind, :] if len(G.coords.shape) else np.ndarray(None)
        Gnew = graphs.Graph(Wnew, coords=coords, lap_type=G.lap_type,
                            plotting=G.plotting)
    else:
        Gnew = Lnew

    return Gnew


def pyramid_analysis(Gs, f, **kwargs):
    r"""Compute the graph pyramid transform coefficients.

    Parameters
    ----------
    Gs : list of graphs
        A multiresolution sequence of graph structures.
    f : ndarray
        Graph signal to analyze.
    h_filters : list
        A list of filter that will be used for the analysis and sythesis operator.
        If only one filter is given, it will be used for all levels.
        Default is h(x) = 1 / (2x+1)

    Returns
    -------
    ca : ndarray
        Coarse approximation at each level
    pe : ndarray
        Prediction error at each level
    h_filters : list
        Graph spectral filters applied

    References
    ----------
    See :cite:`shuman2013framework` and :cite:`pesenson2009variational`.

    """
    if np.shape(f)[0] != Gs[0].N:
        raise ValueError("PYRAMID ANALYSIS: The signal to analyze should have the same dimension as the first graph.")

    levels = len(Gs) - 1

    # check if the type of filters is right.
    h_filters = kwargs.pop('h_filters', lambda x: 1. / (2*x+1))

    if not isinstance(h_filters, list):
        if hasattr(h_filters, '__call__'):
            logger.warning('Converting filters into a list.')
            h_filters = [h_filters]
        else:
            logger.error('Filters must be a list of functions.')

    if len(h_filters) == 1:
        h_filters = h_filters * levels

    elif len(h_filters) != levels:
        message = 'The number of filters must be one or equal to {}.'.format(levels)
        raise ValueError(message)

    ca = [f]
    pe = []

    for i in range(levels):
        # Low pass the signal
        s_low = _analysis(filters.Filter(Gs[i], h_filters[i]), ca[i], **kwargs)
        # Keep only the coefficient on the selected nodes
        ca.append(s_low[Gs[i+1].mr['idx']])
        # Compute prediction
        s_pred = interpolate(Gs[i], ca[i+1], Gs[i+1].mr['idx'], **kwargs)
        # Compute errors
        pe.append(ca[i] - s_pred)

    return ca, pe


def pyramid_synthesis(Gs, cap, pe, order=30, **kwargs):
    r"""Synthesize a signal from its pyramid coefficients.

    Parameters
    ----------
    Gs : Array of Graphs
        A multiresolution sequence of graph structures.
    cap : ndarray
        Coarsest approximation of the original signal.
    pe : ndarray
        Prediction error at each level.
    use_exact : bool
        To use exact graph spectral filtering instead of the Chebyshev approximation.
    order : int
        Degree of the Chebyshev approximation (default=30).
    least_squares : bool
        To use the least squares synthesis (default=False).
    h_filters : ndarray
        The filters used in the analysis operator.
        These are required for least squares synthesis, but not for the direct synthesis method.
    use_landweber : bool
        To use the Landweber iteration approximation in the least squares synthesis.
    reg_eps : float
        Interpolation parameter.
    landweber_its : int
        Number of iterations in the Landweber approximation for least squares synthesis.
    landweber_tau : float
        Parameter for the Landweber iteration.

    Returns
    -------
    reconstruction : ndarray
        The reconstructed signal.
    ca : ndarray
        Coarse approximations at each level

    """
    least_squares = bool(kwargs.pop('least_squares', False))
    def_ul = Gs[0].N > 3000 or Gs[0]._e is None or Gs[0]._U is None
    use_landweber = bool(kwargs.pop('use_landweber', def_ul))
    reg_eps = float(kwargs.get('reg_eps', 0.005))

    if least_squares and 'h_filters' not in kwargs:
        ValueError('h-filters not provided.')

    levels = len(Gs) - 1
    if len(pe) != levels:
        ValueError('Gs and pe have different shapes.')

    ca = [cap]

    # Reconstruct each level
    for i in range(levels):

        if not least_squares:
            s_pred = interpolate(Gs[levels - i - 1], ca[i], Gs[levels - i].mr['idx'],
                                 order=order, reg_eps=reg_eps, **kwargs)
            ca.append(s_pred + pe[levels - i - 1])

        else:
            ca.append(_pyramid_single_interpolation(Gs[levels - i - 1], ca[i],
                      pe[levels - i - 1], h_filters[levels - i - 1],
                      use_landweber=use_landweber, **kwargs))

    ca.reverse()
    reconstruction = ca[0]

    return reconstruction, ca


def _pyramid_single_interpolation(G, ca, pe, keep_inds, h_filter, **kwargs):
    r"""Synthesize a single level of the graph pyramid transform.

    Parameters
    ----------
    G : Graph
        Graph structure on which the signal resides.
    ca : ndarray
        Coarse approximation of the signal on a reduced graph.
    pe : ndarray
        Prediction error that was made when forming the current coarse approximation.
    keep_inds : ndarray
        The indices of the vertices to keep when downsampling the graph and signal.
    h_filter : lambda expression
        The filter in use at this level.
    use_landweber : bool
        To use the Landweber iteration approximation in the least squares synthesis.
        Default is False.
    reg_eps : float
        Interpolation parameter. Default is 0.005.
    landweber_its : int
        Number of iterations in the Landweber approximation for least squares synthesis.
        Default is 50.
    landweber_tau : float
        Parameter for the Landweber iteration. Default is 1.

    Returns
    -------
    finer_approx :
        Coarse approximation of the signal on a higher resolution graph.

    """
    nb_ind = keep_inds.shape
    N = G.N
    reg_eps = float(kwargs.pop('reg_eps', 0.005))
    use_landweber = bool(kwargs.pop('use_landweber', False))
    landweber_its = int(kwargs.pop('landweber_its', 50))
    landweber_tau = float(kwargs.pop('landweber_tau', 1.))

    # index matrix (nb_ind x N) of keep_inds, S_i,j = 1 iff keep_inds[i] = j
    S = sparse.csr_matrix(([1] * nb_ind, (range(nb_ind), keep_inds)), shape=(nb_ind, N))

    if use_landweber:
        x = np.zeros(N)
        z = np.concatenate((ca, pe), axis=0)
        green_kernel = filters.Filter(G, lambda x: 1./(x+reg_eps))
        PhiVlt = _analysis(green_kernel, S.T, **kwargs).T
        filt = filters.Filter(G, h_filter, **kwargs)

        for iteration in range(landweber_its):
            h_filtered_sig = _analysis(filt, x, **kwargs)
            x_bar = h_filtered_sig[keep_inds]
            y_bar = x - interpolate(G, x_bar, keep_inds, **kwargs)
            z_delt = np.concatenate((x_bar, y_bar), axis=0)
            z_delt = z - z_delt
            alpha_new = PhiVlt * z_delt[nb_ind:]
            x_up = sparse.csr_matrix((z_delt, (range(nb_ind), [1] * nb_ind)), shape=(N, 1))
            reg_L = G.L + reg_esp * sparse.eye(N)

            elim_inds = np.setdiff1d(np.arange(N, dtype=int), keep_inds)
            L_red = reg_L[np.ix_(keep_inds, keep_inds)]
            L_in_out = reg_L[np.ix_(keep_inds, elim_inds)]
            L_out_in = reg_L[np.ix_(elim_inds, keep_inds)]
            L_comp = reg_L[np.ix_(elim_inds, elim_inds)]

            next_term = L_red * alpha_new - L_in_out * linalg.spsolve(L_comp, L_out_in * alpha_new)
            next_up = sparse.csr_matrix((next_term, (keep_inds, [1] * nb_ind)), shape=(N, 1))
            x += landweber_tau * _analysis(filt, x_up - next_up, **kwargs) + z_delt[nb_ind:]

        finer_approx = x

    else:
        # When the graph is small enough, we can do a full eigendecomposition
        # and compute the full analysis operator T_a
        H = G.U * sparse.diags(h_filter(G.e), 0) * G.U.T
        Phi = G.U * sparse.diags(1./(reg_eps + G.e), 0) * G.U.T
        Ta = np.concatenate((S * H, sparse.eye(G.N) - Phi[:, keep_inds] * linalg.spsolve(Phi[np.ix_(keep_inds, keep_inds)], S*H)), axis=0)
        finer_approx = linalg.spsolve(Ta.T * Ta, Ta.T * np.concatenate((ca, pe), axis=0))


def _tree_depths(A, root):
    if not graphs.Graph(A=A).is_connected():
        raise ValueError('Graph is not connected')

    N = np.shape(A)[0]
    assigned = root - 1
    depths = np.zeros((N))
    parents = np.zeros((N))

    next_to_expand = np.array([root])
    current_depth = 1

    while len(assigned) < N:
        new_entries_whole_round = []
        for i in range(len(next_to_expand)):
            neighbors = np.where(A[next_to_expand[i]])[0]
            new_entries = np.setdiff1d(neighbors, assigned)
            parents[new_entries] = next_to_expand[i]
            depths[new_entries] = current_depth
            assigned = np.concatenate((assigned, new_entries))
            new_entries_whole_round = np.concatenate((new_entries_whole_round,
                                                      new_entries))

        current_depth = current_depth + 1
        next_to_expand = new_entries_whole_round

    return depths, parents


def tree_multiresolution(G, Nlevel, reduction_method='resistance_distance',
                         compute_full_eigen=False, root=None):
    r"""Compute a multiresolution of trees

    Parameters
    ----------
    G : Graph
        Graph structure of a tree.
    Nlevel : Number of times to downsample and coarsen the tree
    root : int
        The index of the root of the tree. (default = 1)
    reduction_method : str
        The graph reduction method (default = 'resistance_distance')
    compute_full_eigen : bool
        To also compute the graph Laplacian eigenvalues for every tree in the sequence

    Returns
    -------
    Gs : ndarray
        Ndarray, with each element containing a graph structure represent a reduced tree.
    subsampled_vertex_indices : ndarray
        Indices of the vertices of the previous tree that are kept for the subsequent tree.

    """

    if not root:
        if hasattr(G, 'root'):
            root = G.root
        else:
            root = 1

    Gs = [G]

    if compute_full_eigen:
        Gs[0].compute_fourier_basis()

    subsampled_vertex_indices = []
    depths, parents = _tree_depths(G.A, root)
    old_W = G.W

    for lev in range(Nlevel):
        # Identify the vertices in the even depths of the current tree
        down_odd = round(depths) % 2
        down_even = np.ones((Gs[lev].N)) - down_odd
        keep_inds = np.where(down_even == 1)[0]
        subsampled_vertex_indices.append(keep_inds)

        # There will be one undirected edge in the new graph connecting each
        # non-root subsampled vertex to its new parent. Here, we find the new
        # indices of the new parents
        non_root_keep_inds, new_non_root_inds = np.setdiff1d(keep_inds, root)
        old_parents_of_non_root_keep_inds = parents[non_root_keep_inds]
        old_grandparents_of_non_root_keep_inds = parents[old_parents_of_non_root_keep_inds]
        # TODO new_non_root_parents = dsearchn(keep_inds, old_grandparents_of_non_root_keep_inds)

        old_W_i_inds, old_W_j_inds, old_W_weights = sparse.find(old_W)
        i_inds = np.concatenate((new_non_root_inds, new_non_root_parents))
        j_inds = np.concatenate((new_non_root_parents, new_non_root_inds))
        new_N = np.sum(down_even)

        if reduction_method == "unweighted":
            new_weights = np.ones(np.shape(i_inds))

        elif reduction_method == "sum":
            # TODO old_weights_to_parents_inds = dsearchn([old_W_i_inds,old_W_j_inds], [non_root_keep_inds, old_parents_of_non_root_keep_inds]);
            old_weights_to_parents = old_W_weights[old_weights_to_parents_inds]
            # old_W(non_root_keep_inds,old_parents_of_non_root_keep_inds);
            # TODO old_weights_parents_to_grandparents_inds = dsearchn([old_W_i_inds, old_W_j_inds], [old_parents_of_non_root_keep_inds, old_grandparents_of_non_root_keep_inds])
            old_weights_parents_to_grandparents = old_W_weights[old_weights_parents_to_grandparents_inds]
            # old_W(old_parents_of_non_root_keep_inds,old_grandparents_of_non_root_keep_inds);
            new_weights = old_weights_to_parents + old_weights_parents_to_grandparents
            new_weights = np.concatenate((new_weights. new_weights))

        elif reduction_method == "resistance_distance":
            # TODO old_weights_to_parents_inds = dsearchn([old_W_i_inds, old_W_j_inds], [non_root_keep_inds, old_parents_of_non_root_keep_inds])
            old_weights_to_parents = old_W_weight[sold_weights_to_parents_inds]
            # old_W(non_root_keep_inds,old_parents_of_non_root_keep_inds);
            # TODO old_weights_parents_to_grandparents_inds = dsearchn([old_W_i_inds, old_W_j_inds], [old_parents_of_non_root_keep_inds, old_grandparents_of_non_root_keep_inds])
            old_weights_parents_to_grandparents = old_W_weights[old_weights_parents_to_grandparents_inds]
            # old_W(old_parents_of_non_root_keep_inds,old_grandparents_of_non_root_keep_inds);
            new_weights = 1./(1./old_weights_to_parents + 1./old_weights_parents_to_grandparents)
            new_weights = np.concatenate(([new_weights, new_weights]))

        else:
            raise ValueError('Unknown graph reduction method.')

        new_W = sparse.csc_matrix((new_weights, (i_inds, j_inds)),
                                  shape=(new_N, new_N))
        # Update parents
        new_root = np.where(keep_inds == root)[0]
        parents = np.zeros(np.shape(keep_inds)[0], np.shape(keep_inds)[0])
        parents[:new_root - 1, new_root:] = new_non_root_parents

        # Update depths
        depths = depths[keep_inds]
        depths = depths/2.

        # Store new tree
        Gtemp = graphs.Graph(new_W, coords=Gs[lev].coords[keep_inds], limits=G.limits, root=new_root)
        #Gs[lev].copy_graph_attributes(Gtemp, False)

        if compute_full_eigen:
            Gs[lev + 1].compute_fourier_basis()

        # Replace current adjacency matrix and root
        Gs.append(Gtemp)

        old_W = new_W
        root = new_root

    return Gs, subsampled_vertex_indices



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse, linalg

from pygsp2 import utils


logger = utils.build_logger(__name__)


class FourierMixIn(object):

    def _check_fourier_properties(self, name, desc):
        if getattr(self, '_' + name) is None:
            self.logger.warning('The {} G.{} is not available, we need to '
                                'compute the Fourier basis. Explicitly call '
                                'G.compute_fourier_basis() once beforehand '
                                'to suppress the warning.'.format(desc, name))
            self.compute_fourier_basis()
        return getattr(self, '_' + name)

    @property
    def U(self):
        r"""Fourier basis (eigenvectors of the Laplacian).

        Is computed by :meth:`compute_fourier_basis`.
        """
        return self._check_fourier_properties('U', 'Fourier basis')

    @property
    def e(self):
        r"""Eigenvalues of the Laplacian (square of graph frequencies).

        Is computed by :meth:`compute_fourier_basis`.
        """
        return self._check_fourier_properties('e', 'eigenvalues vector')

    @property
    def coherence(self):
        r"""Coherence of the Fourier basis.

        The *mutual coherence* between the basis of Kronecker deltas and the
        basis formed by the eigenvectors of the graph Laplacian is defined as

        .. math:: \mu = \max_{\ell,i} \langle U_\ell, \delta_i \rangle
                      = \max_{\ell,i} | U_{\ell, i} |
                      \in \left[ \frac{1}{\sqrt{N}}, 1 \right],

        where :math:`N` is the number of vertices, :math:`\delta_i \in
        \mathbb{R}^N` denotes the Kronecker delta that is non-zero on vertex
        :math:`v_i`, and :math:`U_\ell \in \mathbb{R}^N` denotes the
        :math:`\ell^\text{th}` eigenvector of the graph Laplacian (i.e., the
        :math:`\ell^\text{th}` Fourier mode).

        The coherence is a measure of the localization of the Fourier modes
        (Laplacian eigenvectors). The larger the value, the more localized the
        eigenvectors can be. The extreme is a node that is disconnected from
        the rest of the graph: an eigenvector will be localized as a Kronecker
        delta there (i.e., :math:`\mu = 1`). In the classical setting, Fourier
        modes (which are complex exponentials) are completely delocalized, and
        the coherence is minimal.

        The value is computed by :meth:`compute_fourier_basis`.

        Examples
        --------

        Delocalized eigenvectors.

        >>> graph = graphs.Path(100)
        >>> graph.compute_fourier_basis()
        >>> minimum = 1 / np.sqrt(graph.n_vertices)
        >>> print('{:.2f} in [{:.2f}, 1]'.format(graph.coherence, minimum))
        0.14 in [0.10, 1]
        >>>
        >>> # Plot some delocalized eigenvectors.
        >>> import matplotlib.pyplot as plt
        >>> graph.set_coordinates('line1D')
        >>> _ = graph.plot(graph.U[:, :5])

        Localized eigenvectors.

        >>> graph = graphs.Sensor(64, seed=10)
        >>> graph.compute_fourier_basis()
        >>> minimum = 1 / np.sqrt(graph.n_vertices)
        >>> print('{:.2f} in [{:.2f}, 1]'.format(graph.coherence, minimum))
        0.84 in [0.12, 1]
        >>>
        >>> # Plot the most localized eigenvector.
        >>> import matplotlib.pyplot as plt
        >>> idx = np.argmax(np.abs(graph.U))
        >>> idx_vertex, idx_fourier = np.unravel_index(idx, graph.U.shape)
        >>> _ = graph.plot(graph.U[:, idx_fourier], highlight=idx_vertex)

        """
        return self._check_fourier_properties('coherence',
                                              'Fourier basis coherence')

    def compute_fourier_basis(self, n_eigenvectors=None):
        r"""Compute the (partial) Fourier basis of the graph (cached).

        The result is cached and accessible by the :attr:`U`, :attr:`e`,
        :attr:`lmax`, and :attr:`coherence` properties.

        Parameters
        ----------
        n_eigenvectors : int or `None`
            Number of eigenvectors to compute. If `None`, all eigenvectors
            are computed. (default: None)

        Notes
        -----
        'G.compute_fourier_basis()' computes a full eigendecomposition of
        the graph Laplacian :math:`L` such that:

        .. math:: L = U \Lambda U^*,

        or a partial eigendecomposition of the graph Laplacian :math:`L`
        such that:

        .. math:: L \approx U \Lambda U^*,

        where :math:`\Lambda` is a diagonal matrix of eigenvalues and the
        columns of :math:`U` are the eigenvectors.

        *G.e* is a vector of length `n_eigenvectors` :math:`\le` *G.N*
        containing the Laplacian eigenvalues. The largest eigenvalue is stored
        in *G.lmax*. The eigenvectors are stored as column vectors of *G.U* in
        the same order that the eigenvalues. Finally, the coherence of the
        Fourier basis is found in *G.coherence*.

        References
        ----------
        See :cite:`chung1997spectral`.

        Examples
        --------
        >>> G = graphs.Torus()
        >>> G.compute_fourier_basis(n_eigenvectors=64)
        >>> G.U.shape
        (256, 64)
        >>> G.e.shape
        (64,)
        >>> G.compute_fourier_basis()
        >>> G.U.shape
        (256, 256)
        >>> G.e.shape
        (256,)
        >>> G.lmax == G.e[-1]
        True
        >>> G.coherence < 1
        True

        """
        if n_eigenvectors is None:
            n_eigenvectors = self.n_vertices

        if (self._U is not None and n_eigenvectors <= len(self._e)):
            return

        assert self.L.shape == (self.n_vertices, self.n_vertices)
        if self.n_vertices**2 * n_eigenvectors > 3000**3:
            self.logger.warning(
                'Computing the {0} eigendecomposition of a large matrix ({1} x'
                ' {1}) is expensive. Consider decreasing n_eigenvectors '
                'or, if using the Fourier basis to filter, using a '
                'polynomial filter instead.'.format(
                    'full' if n_eigenvectors == self.N else 'partial',
                    self.N))

        # TODO: handle non-symmetric Laplacians. Test lap_type?
        if n_eigenvectors == self.n_vertices:
            self._e, self._U = linalg.eigh(self.L.toarray(order='F'), overwrite_a=True)
        else:
            # fast partial eigendecomposition of hermitian matrices
            self._e, self._U = sparse.linalg.eigsh(self.L,
                                                   n_eigenvectors,
                                                   which='SM')
        # Columns are eigenvectors. Sorted in ascending eigenvalue order.

        # Smallest eigenvalue should be zero: correct numerical errors.
        # Eigensolver might sometimes return small negative values, which
        # filter's implementations may not anticipate. Better for plotting too.
        assert -1e-5 < self._e[0] < 1e-5
        self._e[0] = 0

        # Bounded spectrum.
        assert self._e[-1] <= self._get_upper_bound() + 1e-5

        assert np.max(self._e) == self._e[-1]
        if n_eigenvectors == self.N:
            self._lmax = self._e[-1]
            self._lmax_method = 'fourier'
            self._coherence = np.max(np.abs(self._U))

    def gft(self, s):
        r"""Compute the graph Fourier transform.

        The graph Fourier transform of a signal :math:`s` is defined as

        .. math:: \hat{s} = U^* s,

        where :math:`U` is the Fourier basis attr:`U` and :math:`U^*` denotes
        the conjugate transpose or Hermitian transpose of :math:`U`.

        Parameters
        ----------
        s : array_like
            Graph signal in the vertex domain.

        Returns
        -------
        s_hat : ndarray
            Representation of s in the Fourier domain.

        Examples
        --------
        >>> G = graphs.Logo()
        >>> G.compute_fourier_basis()
        >>> s = np.random.default_rng().normal(size=(G.N, 5, 1))
        >>> s_hat = G.gft(s)
        >>> s_star = G.igft(s_hat)
        >>> np.all((s - s_star) < 1e-10)
        True

        """
        s = self._check_signal(s)
        U = np.conjugate(self.U)  # True Hermitian. (Although U is often real.)
        return np.tensordot(U, s, ([0], [0]))

    def igft(self, s_hat):
        r"""Compute the inverse graph Fourier transform.

        The inverse graph Fourier transform of a Fourier domain signal
        :math:`\hat{s}` is defined as

        .. math:: s = U \hat{s},

        where :math:`U` is the Fourier basis :attr:`U`.

        Parameters
        ----------
        s_hat : array_like
            Graph signal in the Fourier domain.

        Returns
        -------
        s : ndarray
            Representation of s_hat in the vertex domain.

        Examples
        --------
        >>> G = graphs.Logo()
        >>> G.compute_fourier_basis()
        >>> s_hat = np.random.default_rng().normal(size=(G.N, 5, 1))
        >>> s = G.igft(s_hat)
        >>> s_hat_star = G.gft(s)
        >>> np.all((s_hat - s_hat_star) < 1e-10)
        True

        """
        s_hat = self._check_signal(s_hat)
        return np.tensordot(self.U, s_hat, ([1], [0]))



r"""
Metro Regression
================

Data imputing example in a  Santiago Metro station.

Compute the number of exits in a random metro station. Then, compute the
regression over the whole network and plot the error of the regression. Here,
Tikhonov regession is used.  Lastly, the average of neighboring nodes is also
used to compare the error of the regression.

To run this example, you need to download three files and place them in the
same directory as this script.

1. Download the file `Tablas de subidas y bajadas nov23.zip` from this link:

https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip

Then, uncompress the zip file and copy `2023.11 Matriz_baj_SS_MH.xlsb` to the
same location as this script.

2. Download the file `santiago_metro_stations_coords.geojson` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson

3. Download the file `santiago_metro_stations_connections.txt` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt
"""
# %%
import sys
import os
import pandas as pd
import numpy as np
from unidecode import unidecode
import networkx as nx
import matplotlib.pyplot as plt
from pygsp2.utils_examples import (
    make_metro_graph, 
    plot_signal_in_graph, 
    metro_database_preprocessing,
    fetch_data
)
from pygsp2 import graphs, learning

current_dir = os.getcwd()
os.chdir(current_dir)

# %% Load data
assets_dir = os.path.join(current_dir, 'data')
fetch_data(assets_dir, "metro")

try:
    commutes = pd.read_excel(
        os.path.join(assets_dir, '2023.11 Matriz_baj_SS_MH.xlsb'),
        header=1,
        sheet_name='bajadas_prom_laboral')
except FileNotFoundError:
    print(f'Data file was not found in:\n {os.getcwd()}')
    print('Download it from:\n' +
          r'https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip')
    sys.exit(1)

# %% Load graph
G, pos = make_metro_graph(
    edgesfile=os.path.join(assets_dir, 'santiago_metro_stations_connections.txt'),
    coordsfile=os.path.join(assets_dir, 'santiago_metro_stations_coords.geojson')
)
pos_list = [(G.nodes[node]['y'], G.nodes[node]['x']) for node in G.nodes]

# Extract adjacency matrix
W = nx.adjacency_matrix(G).toarray()
G_pygsp = graphs.Graph(W)

# Node degree matrix
D = np.diag(G_pygsp.d)
# Compute inverse for later
D_inv = np.linalg.inv(D)

# Convert to uppercase
stations = [name.upper() for name in list(G.nodes)]
stations = [unidecode(station) for station in stations]

# %% Use only metro commutes in database
metro_commutes, signal = metro_database_preprocessing(commutes, stations)

# %% Choose a point in the network and delete its value to regress it

signal2 = signal.copy()
station_idx = np.random.randint(0, len(signal2))
# station_idx = 24

print(f'Deleted Station: {stations[station_idx]}')
signal2[station_idx] = np.nan
mask = np.ones(len(signal)).astype(bool)
mask[station_idx] = False

# Use tikhonov regression to recover the signal
recovered_signal = learning.regression_tikhonov(
    G_pygsp, signal2, mask, tau=0.5)

# Compute the average of the nodes around the missing value
average = (W@D_inv@signal)[station_idx]

print(f'Estimated: {recovered_signal[station_idx]:.2f}')
print(f'One-hop Average: {average:.2f}')
print(f'Real: {signal[station_idx]:.2f}')

# %% Compute error with each station
tikhonov_estimation = np.zeros_like(signal)
average_estimation = W@D_inv@signal

for i, s in enumerate(signal):
    # Allocate new signal
    signal2 = signal.copy()
    print(f'Deleted Station: {stations[i]}')

    # Delete value in the signal
    signal2[i] = np.nan
    mask = np.ones(len(signal)).astype(bool)
    mask[i] = False

    # Use tikhonov regression to recover the signal
    recovered_signal = learning.regression_tikhonov(
        G_pygsp, signal2, mask, tau=0.5)
    tikhonov_estimation[i] = recovered_signal[i]

abs_err = np.abs(tikhonov_estimation - signal)

# %% Plot Tikhonov error in graph
fig, ax = plot_signal_in_graph(G, abs_err,
                     title='Error of Tikhonov Regression',
                     label='Error absoluto')
#fig.savefig('metro_regression_tikhonov_error.png', dpi=300)
# %% Plot Average error in graph
# Change variable to error with average estimation
abs_err = np.abs(average_estimation - signal)

fig, ax = plot_signal_in_graph(G, abs_err,
                     title=r'Error of $y = AD^{-1}x$',
                     label='Error absoluto')
#fig.savefig('metro_regression_error_abs.png', dpi=300)
plt.show()


# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Itersine(Filter):
    r"""Design an itersine filter bank (tight frame).

    Create an itersine half overlap filter bank of Nf filters.
    Going from 0 to lambda_max.

    Parameters
    ----------
    G : graph
    Nf : int (optional)
        Number of filters from 0 to lmax. (default = 6)
    overlap : int (optional)
        (default = 2)

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Itersine(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6, overlap=2):

        self.overlap = overlap
        self.mu = np.linspace(0, G.lmax, num=Nf)

        scales = G.lmax / (Nf - overlap + 1) * overlap

        def kernel(x):
            y = np.cos(x * np.pi)**2
            y = np.sin(0.5 * np.pi * y)
            return y * ((x >= -0.5) * (x <= 0.5))

        kernels = []
        for i in range(1, Nf + 1):

            def kernel_centered(x, i=i):
                y = kernel(x / scales - (i - overlap / 2) / overlap)
                return y * np.sqrt(2 / overlap)

            kernels.append(kernel_centered)

        super(Itersine, self).__init__(G, kernels)

    def _get_extra_repr(self):
        return dict(overlap='{:.2f}'.format(self.overlap))



# -*- coding: utf-8 -*-

from __future__ import division

from functools import partial

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Wave(Filter):
    r"""Design a filter bank of wave kernels.

    The wave kernel is defined in the spectral domain as

    .. math:: g_{\tau, t}(\lambda) = \cos \left( t
        \arccos \left( 1 - \frac{\tau^2}{2} \lambda \right) \right),

    where :math:`\lambda \in [0, 1]` are the normalized eigenvalues of the
    graph Laplacian, :math:`t` is time, and :math:`\tau` is the propagation
    speed.

    The wave kernel is the fundamental solution to the wave equation

    .. math:: - \tau^2 L f(t) = \partial_{tt} f(t),

    where :math:`f: \mathbb{R}_+ \rightarrow \mathbb{R}^N` models, for example,
    the mechanical displacement of a wave on a graph. Given the initial
    condition :math:`f(0)` and assuming a vanishing initial velocity, i.e., the
    first derivative in time of the initial distribution equals zero, the
    solution of the wave equation is expressed as

    .. math:: f(t) = U g_{\tau, t}(\Lambda) U^\top f(0)
                   = g_{\tau, t}(L) f(0).

    The above is, by definition, the convolution of the signal :math:`f(0)`
    with the kernel :math:`g_{\tau, t}`.
    Hence, applying this filter to a signal simulates wave propagation.

    Parameters
    ----------
    G : graph
    time : float or iterable
        Time step.
        If iterable, creates a filter bank with one filter per value.
    speed : float or iterable
        Propagation speed, bounded by 0 (included) and 2 (excluded).
        If iterable, creates a filter bank with one filter per value.

    References
    ----------
    :cite:`grassi2016timevertex`, :cite:`grassi2018timevertex`

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Wave(G, time=[5, 15], speed=1)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    Wave propagation from two sources on a grid.

    >>> import matplotlib.pyplot as plt
    >>> n_side = 11
    >>> graph = graphs.Grid2d(n_side)
    >>> graph.estimate_lmax()
    >>> sources = [
    ...     (n_side//4 * n_side) + (n_side//4),
    ...     (n_side*3//4 * n_side) + (n_side*3//4),
    ... ]
    >>> delta = np.zeros(graph.n_vertices)
    >>> delta[sources] = 5
    >>> steps = np.array([5, 10])
    >>> g = filters.Wave(graph, time=steps, speed=1)
    >>> propagated = g.filter(delta)
    >>> fig, axes = plt.subplots(1, len(steps), figsize=(10, 4))
    >>> _ = fig.suptitle('Wave propagation', fontsize=16)
    >>> for i, ax in enumerate(axes):
    ...     _ = graph.plot(propagated[:, i], highlight=sources,
    ...                    title='step {}'.format(steps[i]), ax=ax)
    ...     ax.set_aspect('equal', 'box')
    ...     ax.set_axis_off()

    """

    def __init__(self, G, time=10, speed=1):

        try:
            iter(time)
        except TypeError:
            time = [time]
        try:
            iter(speed)
        except TypeError:
            speed = [speed]

        self.time = time
        self.speed = speed

        if len(time) != len(speed):
            if len(speed) == 1:
                speed = speed * len(time)
            elif len(time) == 1:
                time = time * len(speed)
            else:
                raise ValueError('If both parameters are iterable, '
                                 'they should have the same length.')

        if np.any(np.asanyarray(speed) >= 2):
            raise ValueError('The wave propagation speed should be in [0, 2[')

        def kernel(x, time, speed):
            return np.cos(time * np.arccos(1 - speed**2 * x / G.lmax / 2))

        kernels = [partial(kernel, time=t, speed=s)
                   for t, s in zip(time, speed)]

        super(Wave, self).__init__(G, kernels)

    def _get_extra_repr(self):
        time = '[' + ', '.join('{:.2f}'.format(t) for t in self.time) + ']'
        speed = '[' + ', '.join('{:.2f}'.format(s) for s in self.speed) + ']'
        return dict(time=time, speed=speed)



r"""
Metro simulation
================

Example of a simulation of the evolution of a graph signal over the Santiago Metro
network.

The initial condition is a graph signal with only one positive integer value
bigger.The plots or animation show how this signal distributes over the network
using:

.. math:: y = AD^-1x

To run this example, you need to download three files and place them in the
same directory as this script.

1. Download the file `Tablas de subidas y bajadas nov23.zip` from this link:

https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip

Then, uncompress the zip file and copy `2023.11 Matriz_baj_SS_MH.xlsb` to the
same location as this script.

2. Download the file `santiago_metro_stations_coords.geojson` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson

3. Download the file `santiago_metro_stations_connections.txt` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt
"""
# %%
import os
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from matplotlib import animation
from pygsp2.utils_examples import make_metro_graph, fetch_data
import pygsp2 as pg

current_dir = os.getcwd()
os.chdir(current_dir)


# If set to true make animation,
# otherwise store each frame as png
MAKE_ANIMATION = True

# %% Download data
assets_dir = os.path.join(current_dir, 'data')
fetch_data(assets_dir, "metro")


# %% Load graph and compute adjacency and node degree
G, pos = make_metro_graph(
    edgesfile=os.path.join(assets_dir, 'santiago_metro_stations_connections.txt'),
    coordsfile=os.path.join(assets_dir, 'santiago_metro_stations_coords.geojson')
)

W = nx.adjacency_matrix(G).toarray()
D = np.diag(W.sum(1))
D_inv = np.linalg.inv(D)

# %% Initialize parameters
NSTEPS = 30  # Arbitrary units, steps
INIT_VALUE = 5000  # Initial conditions
signal = np.zeros([NSTEPS, len(W)])
signal[0, np.random.randint(0, len(W), 1)] = INIT_VALUE

# %% Simulate advancing through the graph
# As a physical distribution network

print('Prerforming simulation...')

for i in np.arange(1, NSTEPS):
    signal[i, :] = W@D_inv@signal[i-1, :]

print('Finished.')

# %% Make animation
if MAKE_ANIMATION:
    # Initiate figure
    fig, ax = plt.subplots(1, 1, figsize=(10, 9))
    cmap = plt.get_cmap('viridis')

    # Draw edges and nodes
    im = nx.draw_networkx_edges(G, pos, node_size=20, ax=ax)
    im = nx.draw_networkx_nodes(G, pos, node_color='gray', node_size=20, ax=ax)
    ax.set_title('T = 0')
    cbar = plt.colorbar(im, ax=ax, ticks=[0, 0.5, 1], label='Signal')
    cbar.set_ticklabels([0, np.amax(signal)/2, np.amax(signal)])

    # Define function for animation

    def update(frame):
        """Function that defines what happens in each frame."""

        idxs = ((np.where(signal[frame, :] > 0)[0]).astype(int))
        nodelist = list(np.array(G)[idxs])
        colors = cmap(signal[frame, idxs] / INIT_VALUE)

        nx.draw_networkx_nodes(
            G, pos, node_color='gray', node_size=20, ax=ax)
        nx.draw_networkx_nodes(G, pos, node_color=colors, nodelist=nodelist,
                               node_size=20, ax=ax)
        ax.set_title(f'T = {frame}')

    anim = animation.FuncAnimation(fig, update,
                                   frames=np.arange(0, len(signal)),
                                   interval=50)

    # saving to gif using ffmpeg writer
    writervideo = animation.PillowWriter(fps=5)

    anim.save('metro_simulation.gif', writer=writervideo)
    plt.show()

else:
    try:
        os.mkdir('metro_simulation/')
    except FileExistsError:
        print(
            'Warning: It seems like this folder already exists. Overwritting...')

    # Initiate figure
    fig, ax = plt.subplots(1, 1, figsize=(10, 9))
    cmap = plt.get_cmap('viridis')

    # Draw edges and nodes
    im = nx.draw_networkx_edges(G, pos, node_size=20, ax=ax)
    im = nx.draw_networkx_nodes(G, pos, node_color='gray', node_size=20, ax=ax)
    ax.set_title('T = 0')
    cbar = plt.colorbar(im, ax=ax, ticks=[0, 0.5, 1], label='Signal')
    cbar.set_ticklabels([0, np.amax(signal)/2, np.amax(signal)])

    # Iterate through each graph signal
    for i, s in enumerate(signal):

        idxs = (np.where(s > 0)[0]).astype(int)
        nodelist = list(np.array(G)[idxs])
        colors = cmap(s[idxs] / INIT_VALUE)

        im = nx.draw_networkx_nodes(
            G, pos, node_color='gray', node_size=20, ax=ax)
        im = nx.draw_networkx_nodes(G, pos, node_color=colors, nodelist=nodelist,
                                    node_size=20, ax=ax)
        ax.set_title(f'T = {i}')

        fig.savefig(f'metro_simulation/{i}.png')



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Simoncelli(Filter):
    r"""Design 2 filters with the Simoncelli construction (tight frame).

    This function creates a Parseval filter bank of 2 filters.
    The low-pass filter is defined by the function

    .. math:: f_{l}=\begin{cases} 1 & \mbox{if }x\leq a\\
            \cos\left(\frac{\pi}{2}\frac{\log\left(\frac{x}{2}\right)}{\log(2)}\right) & \mbox{if }a<x\leq2a\\
            0 & \mbox{if }x>2a \end{cases}

    The high pass filter is adapted to obtain a tight frame.

    Parameters
    ----------
    G : graph
    a : float
        See above equation for this parameter.
        The spectrum is scaled between 0 and 2 (default = 2/3).

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Simoncelli(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, a=2/3):

        self.a = a

        def kernel(x, a):
            y = np.empty(np.shape(x))
            l1 = a
            l2 = 2 * a

            r1ind = (x >= 0) * (x < l1)
            r2ind = (x >= l1) * (x < l2)
            r3ind = (x >= l2)

            y[r1ind] = 1
            y[r2ind] = np.cos(np.pi/2 * np.log(x[r2ind]/a) / np.log(2))
            y[r3ind] = 0

            return y

        simoncelli = Filter(G, lambda x: kernel(x*2/G.lmax, a))
        complement = simoncelli.complement(frame_bound=1)
        kernels = simoncelli._kernels + complement._kernels

        super(Simoncelli, self).__init__(G, kernels)

    def _get_extra_repr(self):
        return dict(a='{:.2f}'.format(self.a))



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2.graphs import NNGraph  # prevent circular import in Python < 3.5


class ImgPatches(NNGraph):
    r"""NN-graph between patches of an image.

    Extract a feature vector in the form of a patch for every pixel of an
    image, then construct a nearest-neighbor graph between these feature
    vectors. The feature matrix, i.e. the patches, can be found in :attr:`Xin`.

    Parameters
    ----------
    img : array
        Input image.
    patch_shape : tuple, optional
        Dimensions of the patch window. Syntax: (height, width), or (height,),
        in which case width = height.
    kwargs : dict
        Parameters passed to :class:`NNGraph`.

    See Also
    --------
    Grid2dImgPatches

    Notes
    -----
    The feature vector of a pixel `i` will consist of the stacking of the
    intensity values of all pixels in the patch centered at `i`, for all color
    channels. So, if the input image has `d` color channels, the dimension of
    the feature vector of each pixel is (patch_shape[0] * patch_shape[1] * d).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from skimage import data, img_as_float
    >>> img = img_as_float(data.camera()[::64, ::64])
    >>> G = graphs.ImgPatches(img, patch_shape=(3, 3))
    >>> print('{} nodes ({} x {} pixels)'.format(G.Xin.shape[0], *img.shape))
    64 nodes (8 x 8 pixels)
    >>> print('{} features per node'.format(G.Xin.shape[1]))
    9 features per node
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, img, patch_shape=(3, 3), **kwargs):

        self.img = img
        self.patch_shape = patch_shape

        try:
            h, w, d = img.shape
        except ValueError:
            try:
                h, w = img.shape
                d = 0
            except ValueError:
                print("Image should be at least a 2D array.")

        try:
            r, c = patch_shape
        except ValueError:
            r = patch_shape[0]
            c = r

        pad_width = [(int((r - 0.5) / 2.), int((r + 0.5) / 2.)),
                     (int((c - 0.5) / 2.), int((c + 0.5) / 2.))]

        if d == 0:
            window_shape = (r, c)
            d = 1  # For the reshape in the return call
        else:
            pad_width += [(0, 0)]
            window_shape = (r, c, d)

        # Pad the image.
        img = np.pad(img, pad_width=pad_width, mode='symmetric')

        # Extract patches as node features.
        # Alternative: sklearn.feature_extraction.image.extract_patches_2d.
        #              sklearn has much less dependencies than skimage.
        try:
            import skimage
        except Exception as e:
            raise ImportError('Cannot import skimage, which is needed to '
                              'extract patches. Try to install it with '
                              'pip (or conda) install scikit-image. '
                              'Original exception: {}'.format(e))
        patches = skimage.util.view_as_windows(img, window_shape=window_shape)
        patches = patches.reshape((h * w, r * c * d))

        super(ImgPatches, self).__init__(patches, **kwargs)

    def _get_extra_repr(self):
        attrs = dict(patch_shape=self.patch_shape)
        attrs.update(super(ImgPatches, self)._get_extra_repr())
        return attrs



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.learning` module provides functions to solve learning problems.

Semi-supervized learning
========================

Those functions help to solve a semi-supervized learning problem, i.e., a
problem where only some values of a graph signal are known and the others shall
be inferred.

.. autosummary::

    regression_tikhonov
    classification_tikhonov
    classification_tikhonov_simplex

"""

import numpy as np
from scipy import sparse


def _import_pyunlocbox():
    try:
        from pyunlocbox import functions, solvers
    except Exception as e:
        raise ImportError('Cannot import pyunlocbox, which is needed to solve '
                          'this optimization problem. Try to install it with '
                          'pip (or conda) install pyunlocbox. '
                          'Original exception: {}'.format(e))
    return functions, solvers


def _to_logits(x):
    logits = np.zeros([len(x), np.max(x)+1])
    logits[range(len(x)), x] = 1
    return logits


def classification_tikhonov_simplex(G, y, M, tau=0.1, **kwargs):
    r"""Solve a classification problem on graph via Tikhonov minimization
    with simple constraints.

    The function first transforms :math:`y` in logits :math:`Y`, then solves

    .. math:: \operatorname*{arg min}_X \| M X - Y \|_2^2 + \tau \ tr(X^T L X)
              \text{ s.t. } sum(X) = 1 \text{ and } X >= 0,

    where :math:`X` and :math:`Y` are logits.

    Parameters
    ----------
    G : :class:`pygsp2.graphs.Graph`
    y : array, length G.n_vertices
        Measurements.
    M : array of boolean, length G.n_vertices
        Masking vector.
    tau : float
        Regularization parameter.
    kwargs : dict
        Parameters for :func:`pyunlocbox.solvers.solve`.

    Returns
    -------
    logits : array, length G.n_vertices
        The logits :math:`X`.

    Examples
    --------
    >>> from pygsp2 import graphs, learning
    >>> import matplotlib.pyplot as plt
    >>>
    >>> G = graphs.Logo()
    >>> G.estimate_lmax()

    Create a ground truth signal:

    >>> signal = np.zeros(G.n_vertices)
    >>> signal[G.info['idx_s']] = 1
    >>> signal[G.info['idx_p']] = 2

    Construct a measurement signal from a binary mask:

    >>> rng = np.random.default_rng(42)
    >>> mask = rng.uniform(0, 1, G.n_vertices) > 0.5
    >>> measures = signal.copy()
    >>> measures[~mask] = np.nan

    Solve the classification problem by reconstructing the signal:

    >>> recovery = learning.classification_tikhonov_simplex(
    ...     G, measures, mask, tau=0.1, verbosity='NONE')

    Plot the results.
    Note that we recover the class with ``np.argmax(recovery, axis=1)``.

    >>> prediction = np.argmax(recovery, axis=1)
    >>> fig, ax = plt.subplots(2, 3, sharey=True, figsize=(10, 6))
    >>> _ = G.plot(signal, ax=ax[0, 0], title='Ground truth')
    >>> _ = G.plot(measures, ax=ax[0, 1], title='Measurements')
    >>> _ = G.plot(prediction, ax=ax[0, 2], title='Recovered class')
    >>> _ = G.plot(recovery[:, 0], ax=ax[1, 0], title='Logit 0')
    >>> _ = G.plot(recovery[:, 1], ax=ax[1, 1], title='Logit 1')
    >>> _ = G.plot(recovery[:, 2], ax=ax[1, 2], title='Logit 2')
    >>> _ = fig.tight_layout()

    """

    functions, solvers = _import_pyunlocbox()

    if tau <= 0:
        raise ValueError('Tau should be greater than 0.')

    y = y.copy()
    y[M == False] = 0
    Y = _to_logits(y.astype(int))
    Y[M == False, :] = 0

    def proj_simplex(y):
        d = y.shape[1]
        a = np.ones(d)
        idx = np.argsort(y)

        def evalpL(y, k, idx):
            return np.sum(y[idx[k:]] - y[idx[k]]) - 1

        def bisectsearch(idx, y):
            idxL, idxH = 0, d-1
            L = evalpL(y, idxL, idx)
            H = evalpL(y, idxH, idx)

            if L < 0:
                return idxL

            while (idxH-idxL) > 1:
                iMid = int((idxL + idxH) / 2)
                M = evalpL(y, iMid, idx)

                if M > 0:
                    idxL, L = iMid, M
                else:
                    idxH, H = iMid, M

            return idxH

        def proj(idx, y):
            k = bisectsearch(idx, y)
            lam = (np.sum(y[idx[k:]]) - 1) / (d - k)
            return np.maximum(0, y - lam)

        x = np.empty_like(y)
        for i in range(len(y)):
            x[i] = proj(idx[i], y[i])
        # x = np.stack(map(proj, idx, y))

        return x

    def smooth_eval(x):
        xTLx = np.sum(x * (G.L.dot(x)))
        e = M * ((M * x.T) - Y.T)
        l2 = np.sum(e * e)
        return tau * xTLx + l2

    def smooth_grad(x):
        return 2 * ((M * (M * x.T - Y.T)).T + tau * G.L * x)

    f1 = functions.func()
    f1._eval = smooth_eval
    f1._grad = smooth_grad

    f2 = functions.func()
    f2._eval = lambda x: 0  # Indicator functions evaluate to zero.
    f2._prox = lambda x, step: proj_simplex(x)

    step = 0.5 / (1 + tau * G.lmax)
    solver = solvers.forward_backward(step=step)
    ret = solvers.solve([f1, f2], Y.copy(), solver, **kwargs)
    return ret['sol']


def classification_tikhonov(G, y, M, tau=0):
    r"""Solve a classification problem on graph via Tikhonov minimization.

    The function first transforms :math:`y` in logits :math:`Y`, then solves

    .. math:: \operatorname*{arg min}_X \| M X - Y \|_2^2 + \tau \ tr(X^T L X)

    if :math:`\tau > 0`, and

    .. math:: \operatorname*{arg min}_X tr(X^T L X) \ \text{ s. t. } \ Y = M X

    otherwise, where :math:`X` and :math:`Y` are logits.
    The function returns the maximum of the logits.

    Parameters
    ----------
    G : :class:`pygsp2.graphs.Graph`
    y : array, length G.n_vertices
        Measurements.
    M : array of boolean, length G.n_vertices
        Masking vector.
    tau : float
        Regularization parameter.

    Returns
    -------
    logits : array, length G.n_vertices
        The logits :math:`X`.

    Examples
    --------
    >>> from pygsp2 import graphs, learning
    >>> import matplotlib.pyplot as plt
    >>>
    >>> G = graphs.Logo()

    Create a ground truth signal:

    >>> signal = np.zeros(G.n_vertices)
    >>> signal[G.info['idx_s']] = 1
    >>> signal[G.info['idx_p']] = 2

    Construct a measurement signal from a binary mask:

    >>> rng = np.random.default_rng(42)
    >>> mask = rng.uniform(0, 1, G.n_vertices) > 0.5
    >>> measures = signal.copy()
    >>> measures[~mask] = np.nan

    Solve the classification problem by reconstructing the signal:

    >>> recovery = learning.classification_tikhonov(G, measures, mask, tau=0)

    Plot the results.
    Note that we recover the class with ``np.argmax(recovery, axis=1)``.

    >>> prediction = np.argmax(recovery, axis=1)
    >>> fig, ax = plt.subplots(2, 3, sharey=True, figsize=(10, 6))
    >>> _ = G.plot(signal, ax=ax[0, 0], title='Ground truth')
    >>> _ = G.plot(measures, ax=ax[0, 1], title='Measurements')
    >>> _ = G.plot(prediction, ax=ax[0, 2], title='Recovered class')
    >>> _ = G.plot(recovery[:, 0], ax=ax[1, 0], title='Logit 0')
    >>> _ = G.plot(recovery[:, 1], ax=ax[1, 1], title='Logit 1')
    >>> _ = G.plot(recovery[:, 2], ax=ax[1, 2], title='Logit 2')
    >>> _ = fig.tight_layout()

    """
    y = y.copy()
    y[M == False] = 0
    Y = _to_logits(y.astype(int))
    return regression_tikhonov(G, Y, M, tau)


def regression_tikhonov(G, y, M, tau=0):
    r"""Solve a regression problem on graph via Tikhonov minimization.

    The function solves

    .. math:: \operatorname*{arg min}_x \| M x - y \|_2^2 + \tau \ x^T L x

    if :math:`\tau > 0`, and

    .. math:: \operatorname*{arg min}_x x^T L x \ \text{ s. t. } \ y = M x

    otherwise.

    Parameters
    ----------
    G : :class:`pygsp2.graphs.Graph`
    y : array, length G.n_vertices
        Measurements.
    M : array of boolean, length G.n_vertices
        Masking vector.
    tau : float
        Regularization parameter.

    Returns
    -------
    x : array, length G.n_vertices
        Recovered values :math:`x`.

    Examples
    --------
    >>> from pygsp2 import graphs, filters, learning
    >>> import matplotlib.pyplot as plt
    >>>
    >>> G = graphs.Sensor(N=100, seed=42)
    >>> G.estimate_lmax()

    Create a smooth ground truth signal:

    >>> filt = lambda x: 1 / (1 + 10*x)
    >>> filt = filters.Filter(G, filt)
    >>> rng = np.random.default_rng(42)
    >>> signal = filt.analyze(rng.normal(size=G.n_vertices))

    Construct a measurement signal from a binary mask:

    >>> mask = rng.uniform(0, 1, G.n_vertices) > 0.5
    >>> measures = signal.copy()
    >>> measures[~mask] = np.nan

    Solve the regression problem by reconstructing the signal:

    >>> recovery = learning.regression_tikhonov(G, measures, mask, tau=0)

    Plot the results:

    >>> fig, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(10, 3))
    >>> limits = [signal.min(), signal.max()]
    >>> _ = G.plot(signal, ax=ax1, limits=limits, title='Ground truth')
    >>> _ = G.plot(measures, ax=ax2, limits=limits, title='Measures')
    >>> _ = G.plot(recovery, ax=ax3, limits=limits, title='Recovery')
    >>> _ = fig.tight_layout()

    """

    if tau > 0:
        y = y.copy()
        y[M == False] = 0

        if sparse.issparse(G.L):

            def Op(x):
                return (M * x.T).T + tau * (G.L.dot(x))

            LinearOp = sparse.linalg.LinearOperator([G.N, G.N], Op)

            if y.ndim > 1:
                sol = np.empty(shape=y.shape)
                res = np.empty(shape=y.shape[1])
                for i in range(y.shape[1]):
                    sol[:, i], res[i] = sparse.linalg.cg(
                        LinearOp, y[:, i])
            else:
                sol, res = sparse.linalg.cg(LinearOp, y)

            # TODO: do something with the residual...
            return sol

        else:

            # Creating this matrix may be problematic in term of memory.
            # Consider using an operator instead...
            if type(G.L).__module__ == np.__name__:
                LinearOp = np.diag(M*1) + tau * G.L
            return np.linalg.solve(LinearOp, M * y)

    else:

        if np.prod(M.shape) != G.n_vertices:
            raise ValueError("M should be of size [G.n_vertices,]")

        indl = M
        indu = (M == False)

        Luu = G.L[indu, :][:, indu]
        Wul = - G.L[indu, :][:, indl]

        if sparse.issparse(G.L):
            sol_part = sparse.linalg.spsolve(Luu, Wul.dot(y[indl]))
        else:
            sol_part = np.linalg.solve(Luu, np.matmul(Wul, y[indl]))

        sol = y.copy()
        sol[indu] = sol_part

        return sol



# -*- coding: utf-8 -*-

from __future__ import division

import collections
import copy

import numpy as np
from scipy import sparse, spatial

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class Community(Graph):
    r"""Community graph.

    Parameters
    ----------
    N : int
        Number of nodes (default = 256).
    Nc : int (optional)
        Number of communities (default = :math:`\lfloor \sqrt{N}/2 \rceil`).
    min_comm : int (optional)
        Minimum size of the communities
        (default = :math:`\lfloor N/Nc/3 \rceil`).
    min_deg : int (optional)
        NOT IMPLEMENTED. Minimum degree of each node (default = 0).
    comm_sizes : int (optional)
        Size of the communities (default = random).
    size_ratio : float (optional)
        Ratio between the radius of world and the radius of communities
        (default = 1).
    world_density : float (optional)
        Probability of a random edge between two different communities
        (default = 1/N).
    comm_density : float (optional)
        Probability of a random edge inside any community (default = None,
        which implies k_neigh or epsilon will be used to determine
        intra-edges).
    k_neigh : int (optional)
        Number of intra-community connections.
        Not used if comm_density is defined (default = None, which implies
        comm_density or epsilon will be used to determine intra-edges).
    epsilon : float (optional)
        Largest distance at which two nodes sharing a community are connected.
        Not used if k_neigh or comm_density is defined
        (default = :math:`\sqrt{2\sqrt{N}}/2`).
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Community(N=250, Nc=3, comm_sizes=[50, 120, 80], seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.5)
    >>> _ = G.plot(ax=axes[1])

    """
    def __init__(self,
                 N=256,
                 Nc=None,
                 min_comm=None,
                 min_deg=None,
                 comm_sizes=None,
                 size_ratio=1,
                 world_density=None,
                 comm_density=None,
                 k_neigh=None,
                 epsilon=None,
                 seed=None,
                 **kwargs):

        if Nc is None:
            Nc = int(round(np.sqrt(N) / 2))
        if min_comm is None:
            min_comm = int(round(N / (3 * Nc)))
        if min_deg is not None:
            raise NotImplementedError
        if world_density is None:
            world_density = 1 / N
        if not 0 <= world_density <= 1:
            raise ValueError('World density should be in [0, 1].')
        if epsilon is None:
            epsilon = np.sqrt(2 * np.sqrt(N)) / 2

        self.Nc = Nc
        self.min_comm = min_comm
        self.comm_sizes = comm_sizes
        self.size_ratio = size_ratio
        self.world_density = world_density
        self.comm_density = comm_density
        self.k_neigh = k_neigh
        self.epsilon = epsilon
        self.seed = seed

        rng = np.random.default_rng(seed)

        self.logger = utils.build_logger(__name__)
        w_data = [[], [[], []]]

        if min_comm * Nc > N:
            raise ValueError('The constraint on minimum size for communities is unsolvable.')

        info = {'node_com': None, 'comm_sizes': None, 'world_rad': None,
                'world_density': world_density, 'min_comm': min_comm}

        # Communities construction #
        if comm_sizes is None:
            mandatory_labels = np.tile(np.arange(Nc), (min_comm,))  # min_comm labels for each of the Nc communities
            remaining_labels = rng.choice(Nc, N - min_comm * Nc)  # random choice for the remaining labels
            info['node_com'] = np.sort(np.concatenate((mandatory_labels, remaining_labels)))
        else:
            if len(comm_sizes) != Nc:
                raise ValueError('There should be Nc community sizes.')
            if np.sum(comm_sizes) != N:
                raise ValueError('The sum of community sizes should be N.')
            # create labels based on the constraint given for the community sizes. No random assignation here.
            info['node_com'] = np.concatenate([[val] * cnt for (val, cnt) in enumerate(comm_sizes)])

        counts = collections.Counter(info['node_com'])
        info['comm_sizes'] = np.array([cnt[1] for cnt in sorted(counts.items())])
        info['world_rad'] = size_ratio * np.sqrt(N)

        # Intra-community edges construction #
        if comm_density is not None:
            # random picking edges following the community density (same for all communities)
            if not 0 <= comm_density <= 1:
                raise ValueError('comm_density should be between 0 and 1.')
            info['comm_density'] = comm_density
            self.logger.info('Constructed using community density = {}'.format(comm_density))
        elif k_neigh is not None:
            # k-NN among the nodes in the same community (same k for all communities)
            if k_neigh < 0:
                raise ValueError('k_neigh cannot be negative.')
            info['k_neigh'] = k_neigh
            self.logger.info('Constructed using K-NN with k = {}'.format(k_neigh))
        else:
            # epsilon-NN among the nodes in the same community (same eps for all communities)
            info['epsilon'] = epsilon
            self.logger.info('Constructed using eps-NN with eps = {}'.format(epsilon))

        # Coordinates #
        info['com_coords'] = info['world_rad'] * np.array(list(zip(
            np.cos(2 * np.pi * np.arange(1, Nc + 1) / Nc),
            np.sin(2 * np.pi * np.arange(1, Nc + 1) / Nc))))

        coords = rng.uniform(size=(N, 2))  # nodes' coordinates inside the community
        coords = np.array([[elem[0] * np.cos(2 * np.pi * elem[1]),
                            elem[0] * np.sin(2 * np.pi * elem[1])] for elem in coords])

        for i in range(N):
            # set coordinates as an offset from the center of the community it belongs to
            comm_idx = info['node_com'][i]
            comm_rad = np.sqrt(info['comm_sizes'][comm_idx])
            coords[i] = info['com_coords'][comm_idx] + comm_rad * coords[i]

        first_node = 0
        for i in range(Nc):
            com_siz = info['comm_sizes'][i]
            M = com_siz * (com_siz - 1) / 2

            if comm_density is not None:
                nb_edges = int(comm_density * M)
                tril_ind = np.tril_indices(com_siz, -1)
                indices = rng.permutation(int(M))[:nb_edges]

                w_data[0] += [1] * nb_edges
                w_data[1][0] += [first_node + tril_ind[1][elem] for elem in indices]
                w_data[1][1] += [first_node + tril_ind[0][elem] for elem in indices]

            elif k_neigh is not None:
                comm_coords = coords[first_node:first_node + com_siz]
                kdtree = spatial.KDTree(comm_coords)
                __, indices = kdtree.query(comm_coords, k=k_neigh + 1)

                pairs_set = set()
                map(lambda row: map(lambda elm: pairs_set.add((min(row[0], elm), max(row[0], elm))), row[1:]), indices)

                w_data[0] += [1] * len(pairs_set)
                w_data[1][0] += [first_node + pair[0] for pair in pairs_set]
                w_data[1][1] += [first_node + pair[1] for pair in pairs_set]

            else:
                comm_coords = coords[first_node:first_node + com_siz]
                kdtree = spatial.KDTree(comm_coords)
                pairs_set = kdtree.query_pairs(epsilon)

                w_data[0] += [1] * len(pairs_set)
                w_data[1][0] += [first_node + elem[0] for elem in pairs_set]
                w_data[1][1] += [first_node + elem[1] for elem in pairs_set]

            first_node += com_siz

        # Inter-community edges construction #
        M = (N**2 - np.sum([com_siz**2 for com_siz in info['comm_sizes']])) / 2
        nb_edges = int(world_density * M)

        if world_density < 0.35:
            # use regression sampling
            inter_edges = set()
            while len(inter_edges) < nb_edges:
                new_point = rng.integers(0, N, 2)
                if info['node_com'][min(new_point)] != info['node_com'][max(new_point)]:
                    inter_edges.add((min(new_point), max(new_point)))
        else:
            # use random permutation
            indices = rng.permutation(int(M))[:nb_edges]
            all_points, first_col = [], 0
            for i in range(Nc - 1):
                nb_col = info['comm_sizes'][i]
                first_row = np.sum(info['comm_sizes'][:i+1])

                for j in range(i+1, Nc):
                    nb_row = info['comm_sizes'][j]
                    all_points += [(first_row + r, first_col + c) for r in range(nb_row) for c in range(nb_col)]

                    first_row += nb_row
                first_col += nb_col

            inter_edges = np.array(all_points)[indices]

        w_data[0] += [1] * nb_edges
        w_data[1][0] += [elem[0] for elem in inter_edges]
        w_data[1][1] += [elem[1] for elem in inter_edges]

        w_data[0] += w_data[0]
        tmp_w_data = copy.deepcopy(w_data[1][0])
        w_data[1][0] += w_data[1][1]
        w_data[1][1] += tmp_w_data
        w_data[1] = tuple(w_data[1])

        W = sparse.coo_matrix(tuple(w_data), shape=(N, N))

        for key, value in {'Nc': Nc, 'info': info}.items():
            setattr(self, key, value)

        super(Community, self).__init__(W, coords=coords, **kwargs)

    def _get_extra_repr(self):
        attrs = {'Nc': self.Nc,
                 'min_comm': self.min_comm,
                 'comm_sizes': self.comm_sizes,
                 'size_ratio': '{:.2f}'.format(self.size_ratio),
                 'world_density': '{:.2f}'.format(self.world_density)}
        if self.comm_density is not None:
            attrs['comm_density'] = '{:.2f}'.format(self.comm_density)
        elif self.k_neigh is not None:
            attrs['k_neigh'] = self.k_neigh
        else:
            attrs['epsilon'] = '{:.2f}'.format(self.epsilon)
        attrs['seed'] = self.seed
        return attrs



# -*- coding: utf-8 -*-

"""
Test suite for the filters module of the pygsp2 package.

"""

import unittest
import sys

import numpy as np

from pygsp2 import graphs, filters


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._G = graphs.Sensor(123, seed=42)
        cls._G.compute_fourier_basis()
        cls._rng = np.random.default_rng(42)
        cls._signal = cls._rng.uniform(size=cls._G.N)

    @classmethod
    def tearDownClass(cls):
        pass

    def _generate_coefficients(self, N, Nf, vertex_delta=83):
        S = np.zeros((N*Nf, Nf))
        S[vertex_delta] = 1
        for i in range(Nf):
            S[vertex_delta + i * self._G.N, i] = 1
        return S

    def _test_methods(self, f, tight, check=True):
        self.assertIs(f.G, self._G)

        f.evaluate(self._G.e)
        f.evaluate(np.random.default_rng().uniform(0, 1, size=(4, 6, 3)))

        A, B = f.estimate_frame_bounds(self._G.e)
        if tight:
            np.testing.assert_allclose(A, B)
        else:
            assert B - A > 0.01

        # Analysis.
        s2 = f.filter(self._signal, method='exact')
        s3 = f.filter(self._signal, method='chebyshev', order=100)

        # Synthesis.
        s4 = f.filter(s2, method='exact')
        s5 = f.filter(s3, method='chebyshev', order=100)

        if check:
            # Chebyshev should be close to exact.
            # Does not pass for Gabor, Modulation and Rectangular (not smooth).
            np.testing.assert_allclose(s2, s3, rtol=0.1, atol=0.01)
            np.testing.assert_allclose(s4, s5, rtol=0.1, atol=0.01)

        if tight:
            # Tight frames should not loose information.
            np.testing.assert_allclose(s4, A * self._signal)
            assert np.linalg.norm(s5 - A * self._signal) < 0.1

        # Computing the frame is an alternative way to filter.
        if f.n_filters != self._G.n_vertices:
            # It doesn't work (yet) if the number of filters is the same as the
            # number of nodes as f.filter() infers that we want synthesis where
            # we actually want analysis.
            F = f.compute_frame(method='exact')
            s = F.dot(self._signal).reshape(-1, self._G.N).T.squeeze()
            np.testing.assert_allclose(s, s2)

            F = f.compute_frame(method='chebyshev', order=100)
            s = F.dot(self._signal).reshape(-1, self._G.N).T.squeeze()
            np.testing.assert_allclose(s, s3)

    def test_filter(self, n_filters=5):
        g = filters.MexicanHat(self._G, n_filters)

        s1 = self._rng.uniform(size=self._G.N)
        s2 = s1.reshape((self._G.N, 1))
        s3 = g.filter(s1)
        s4 = g.filter(s2)
        s5 = g.analyze(s1)
        assert s3.shape == (self._G.N, n_filters)
        np.testing.assert_allclose(s3, s4)
        np.testing.assert_allclose(s3, s5)

        s1 = self._rng.uniform(size=(self._G.N, 4))
        s2 = s1.reshape((self._G.N, 4, 1))
        s3 = g.filter(s1)
        s4 = g.filter(s2)
        s5 = g.analyze(s1)
        assert s3.shape == (self._G.N, 4, n_filters)
        np.testing.assert_allclose(s3, s4)
        np.testing.assert_allclose(s3, s5)

        s1 = self._rng.uniform(size=(self._G.N, n_filters))
        s2 = s1.reshape((self._G.N, 1, n_filters))
        s3 = g.filter(s1)
        s4 = g.filter(s2)
        s5 = g.synthesize(s1)
        assert s3.shape == (self._G.N,)
        np.testing.assert_allclose(s3, s4)
        np.testing.assert_allclose(s3, s5)

        s1 = self._rng.uniform(size=(self._G.N, 10, n_filters))
        s3 = g.filter(s1)
        s5 = g.synthesize(s1)
        assert s3.shape == (self._G.N, 10)
        np.testing.assert_allclose(s3, s5)

    def test_localize(self):
        g = filters.Heat(self._G, 100)

        # Localize signal at node by filtering Kronecker delta.
        NODE = 10
        s1 = g.localize(NODE, method='exact')

        # Should be equal to a row / column of the filtering operator.
        gL = self._G.U.dot(np.diag(g.evaluate(self._G.e)[0]).dot(self._G.U.T))
        s2 = np.sqrt(self._G.N) * gL[NODE, :]
        np.testing.assert_allclose(s1, s2)

        # That is actually a row / column of the analysis operator.
        F = g.compute_frame(method='exact')
        np.testing.assert_allclose(F, gL)

    def test_frame_bounds(self):
        # Not a frame, it as a null-space.
        g = filters.Rectangular(self._G)
        A, B = g.estimate_frame_bounds()
        self.assertEqual(A, 0)
        self.assertEqual(B, 1)
        # Identity is tight.
        g = filters.Filter(self._G, lambda x: np.full_like(x, 2))
        A, B = g.estimate_frame_bounds()
        self.assertEqual(A, 4)
        self.assertEqual(B, 4)

    def test_frame(self):
        """The frame is a stack of functions of the Laplacian."""
        g = filters.Heat(self._G, scale=[8, 9])
        gL1 = g.compute_frame(method='exact')
        gL2 = g.compute_frame(method='chebyshev', order=30)
        def get_frame(freq_response):
            return self._G.U.dot(np.diag(freq_response).dot(self._G.U.T))
        gL = np.concatenate([get_frame(gl) for gl in g.evaluate(self._G.e)])
        np.testing.assert_allclose(gL1, gL)
        np.testing.assert_allclose(gL2, gL, atol=1e-10)

    def test_complement(self, frame_bound=2.5):
        """Any filter bank becomes tight upon addition of their complement."""
        g = filters.MexicanHat(self._G)
        g += g.complement(frame_bound)
        A, B = g.estimate_frame_bounds()
        np.testing.assert_allclose(A, frame_bound)
        np.testing.assert_allclose(B, frame_bound)

    def test_inverse(self, frame_bound=3):
        """The frame is the pseudo-inverse of the original frame."""
        g = filters.Heat(self._G, scale=[2, 3, 4])
        h = g.inverse()
        Ag, Bg = g.estimate_frame_bounds()
        Ah, Bh = h.estimate_frame_bounds()
        np.testing.assert_allclose(Ag * Bh, 1)
        np.testing.assert_allclose(Bg * Ah, 1)
        gL = g.compute_frame(method='exact')
        hL = h.compute_frame(method='exact')
        I = np.identity(self._G.N)
        np.testing.assert_allclose(hL.T.dot(gL), I, atol=1e-10)
        pinv = np.linalg.inv(gL.T.dot(gL)).dot(gL.T)
        np.testing.assert_allclose(pinv, hL.T, atol=1e-10)
        # The reconstruction is exact for any frame (lower bound A > 0).
        y = g.filter(self._signal, method='exact')
        z = h.filter(y, method='exact')
        np.testing.assert_allclose(z, self._signal)
        # Not invertible if not a frame.
        g = filters.Expwin(self._G)
        with self.assertLogs(level='WARNING'):
            h = g.inverse()
            h.evaluate(self._G.e)
        # If the frame is tight, inverse is h=g/A.
        g += g.complement(frame_bound)
        h = g.inverse()
        he = g(self._G.e) / frame_bound
        np.testing.assert_allclose(h(self._G.e), he, atol=1e-10)

    def test_custom_filter(self):
        def kernel(x):
            return x / (1. + x)
        f = filters.Filter(self._G, kernels=kernel)
        self.assertEqual(f.Nf, 1)
        self.assertIs(f._kernels[0], kernel)
        self._test_methods(f, tight=False)

    def test_abspline(self):
        f = filters.Abspline(self._G, Nf=4)
        self._test_methods(f, tight=False)

    def test_gabor(self):
        f = filters.Rectangular(self._G, None, 0.1)
        f = filters.Gabor(self._G, f)
        self._test_methods(f, tight=False, check=False)
        self.assertRaises(ValueError, filters.Gabor, graphs.Sensor(), f)
        f = filters.Regular(self._G)
        self.assertRaises(ValueError, filters.Gabor, self._G, f)

    def test_modulation(self):
        f = filters.Rectangular(self._G, None, 0.1)
        # TODO: synthesis doesn't work yet.
        # f = filters.Modulation(self._G, f, modulation_first=False)
        # self._test_methods(f, tight=False, check=False)
        f = filters.Modulation(self._G, f, modulation_first=True)
        self._test_methods(f, tight=False, check=False)
        self.assertRaises(ValueError, filters.Modulation, graphs.Sensor(), f)
        f = filters.Regular(self._G)
        self.assertRaises(ValueError, filters.Modulation, self._G, f)

    def test_modulation_gabor(self):
        """Both should be equivalent for deltas centered at the eigenvalues."""
        f = filters.Rectangular(self._G, 0, 0)
        f1 = filters.Modulation(self._G, f, modulation_first=True)
        f2 = filters.Gabor(self._G, f)
        s1 = f1.filter(self._signal)
        s2 = f2.filter(self._signal)
        np.testing.assert_allclose(abs(s1), abs(s2), atol=1e-5)

    def test_halfcosine(self):
        f = filters.HalfCosine(self._G, Nf=4)
        self._test_methods(f, tight=True)

    def test_itersine(self):
        f = filters.Itersine(self._G, Nf=4)
        self._test_methods(f, tight=True)

    def test_mexicanhat(self):
        f = filters.MexicanHat(self._G, Nf=5, normalize=False)
        self._test_methods(f, tight=False)
        f = filters.MexicanHat(self._G, Nf=4, normalize=True)
        self._test_methods(f, tight=False)

    def test_meyer(self):
        f = filters.Meyer(self._G, Nf=4)
        self._test_methods(f, tight=True)

    def test_simpletf(self):
        f = filters.SimpleTight(self._G, Nf=4)
        self._test_methods(f, tight=True)

    def test_regular(self):
        f = filters.Regular(self._G)
        self._test_methods(f, tight=True)
        f = filters.Regular(self._G, degree=5)
        self._test_methods(f, tight=True)
        f = filters.Regular(self._G, degree=0)
        self._test_methods(f, tight=True)

    def test_held(self):
        f = filters.Held(self._G)
        self._test_methods(f, tight=True)
        f = filters.Held(self._G, a=0.25)
        self._test_methods(f, tight=True)

    def test_simoncelli(self):
        f = filters.Simoncelli(self._G)
        self._test_methods(f, tight=True)
        f = filters.Simoncelli(self._G, a=0.25)
        self._test_methods(f, tight=True)

    def test_papadakis(self):
        f = filters.Papadakis(self._G)
        self._test_methods(f, tight=True)
        f = filters.Papadakis(self._G, a=0.25)
        self._test_methods(f, tight=True)

    def test_heat(self):
        f = filters.Heat(self._G, normalize=False, scale=10)
        self._test_methods(f, tight=False)
        f = filters.Heat(self._G, normalize=False, scale=np.array([5, 10]))
        self._test_methods(f, tight=False)
        f = filters.Heat(self._G, normalize=True, scale=10)
        np.testing.assert_allclose(np.linalg.norm(f.evaluate(self._G.e)), 1)
        self._test_methods(f, tight=False)
        f = filters.Heat(self._G, normalize=True, scale=[5, 10])
        np.testing.assert_allclose(np.linalg.norm(f.evaluate(self._G.e)[0]), 1)
        np.testing.assert_allclose(np.linalg.norm(f.evaluate(self._G.e)[1]), 1)
        self._test_methods(f, tight=False)

    def test_wave(self):
        f = filters.Wave(self._G)
        self._test_methods(f, tight=False)
        f = filters.Wave(self._G, time=1)
        self._test_methods(f, tight=False)
        f = filters.Wave(self._G, time=[1, 2, 3])
        self._test_methods(f, tight=False)
        f = filters.Wave(self._G, speed=[1])
        self._test_methods(f, tight=False)
        f = filters.Wave(self._G, speed=[0.5, 1, 1.5])
        self._test_methods(f, tight=False)
        f = filters.Wave(self._G, time=[1, 2], speed=[1, 1.5])
        self._test_methods(f, tight=False)
        # Sequences of differing lengths.
        self.assertRaises(ValueError, filters.Wave, self._G, time=[1, 2, 3],
                          speed=[0, 1])
        # Invalid speed.
        self.assertRaises(ValueError, filters.Wave, self._G, speed=2)

    def test_expwin(self):
        f = filters.Expwin(self._G)
        self._test_methods(f, tight=False)
        f = filters.Expwin(self._G, band_min=None, band_max=0.8)
        self._test_methods(f, tight=False)
        f = filters.Expwin(self._G, band_min=0.1, band_max=None)
        self._test_methods(f, tight=False)
        f = filters.Expwin(self._G, band_min=0.1, band_max=0.7)
        self._test_methods(f, tight=False)
        f = filters.Expwin(self._G, band_min=None, band_max=None)
        self._test_methods(f, tight=True)

    def test_rectangular(self):
        f = filters.Rectangular(self._G)
        self._test_methods(f, tight=False, check=False)
        f = filters.Rectangular(self._G, band_min=None, band_max=0.8)
        self._test_methods(f, tight=False, check=False)
        f = filters.Rectangular(self._G, band_min=0.1, band_max=None)
        self._test_methods(f, tight=False, check=False)
        f = filters.Rectangular(self._G, band_min=0.1, band_max=0.7)
        self._test_methods(f, tight=False, check=False)
        f = filters.Rectangular(self._G, band_min=None, band_max=None)
        self._test_methods(f, tight=True, check=True)

    def test_approximations(self):
        r"""
        Test that the different methods for filter analysis, i.e. 'exact',
        'cheby', and 'lanczos', produce the same output.
        """
        # TODO: done in _test_methods.

        f = filters.Heat(self._G)
        c_exact = f.filter(self._signal, method='exact')
        c_cheby = f.filter(self._signal, method='chebyshev')

        np.testing.assert_allclose(c_exact, c_cheby)
        self.assertRaises(ValueError, f.filter, self._signal, method='lanczos')


suite = unittest.TestLoader().loadTestsFromTestCase(TestCase)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils


_logger = utils.build_logger(__name__)


@utils.filterbank_handler
def compute_cheby_coeff(f, m=30, N=None, *args, **kwargs):
    r"""
    Compute Chebyshev coefficients for a Filterbank.

    Parameters
    ----------
    f : Filter
        Filterbank with at least 1 filter
    m : int
        Maximum order of Chebyshev coeff to compute
        (default = 30)
    N : int
        Grid order used to compute quadrature
        (default = m + 1)
    i : int
        Index of the Filterbank element to compute
        (default = 0)

    Returns
    -------
    c : ndarray
        Matrix of Chebyshev coefficients

    """
    G = f.G
    i = kwargs.pop('i', 0)

    if not N:
        N = m + 1

    a_arange = [0, G.lmax]

    a1 = (a_arange[1] - a_arange[0]) / 2
    a2 = (a_arange[1] + a_arange[0]) / 2
    c = np.zeros(m + 1)

    tmpN = np.arange(N)
    num = np.cos(np.pi * (tmpN + 0.5) / N)
    for o in range(m + 1):
        c[o] = 2. / N * np.dot(f._kernels[i](a1 * num + a2),
                               np.cos(np.pi * o * (tmpN + 0.5) / N))

    return c


def cheby_op(G, c, signal, **kwargs):
    r"""
    Chebyshev polynomial of graph Laplacian applied to vector.

    Parameters
    ----------
    G : Graph
    c : ndarray or list of ndarrays
        Chebyshev coefficients for a Filter or a Filterbank
    signal : ndarray
        Signal to filter

    Returns
    -------
    r : ndarray
        Result of the filtering

    """
    # Handle if we do not have a list of filters but only a simple filter in cheby_coeff.
    if not isinstance(c, np.ndarray):
        c = np.array(c)

    c = np.atleast_2d(c)
    Nscales, M = c.shape

    if M < 2:
        raise TypeError("The coefficients have an invalid shape")

    # thanks to that, we can also have 1d signal.
    try:
        Nv = np.shape(signal)[1]
        r = np.zeros((G.N * Nscales, Nv))
    except IndexError:
        r = np.zeros((G.N * Nscales))

    a_arange = [0, G.lmax]

    a1 = float(a_arange[1] - a_arange[0]) / 2.
    a2 = float(a_arange[1] + a_arange[0]) / 2.

    twf_old = signal
    twf_cur = (G.L.dot(signal) - a2 * signal) / a1

    tmpN = np.arange(G.N, dtype=int)
    for i in range(Nscales):
        r[tmpN + G.N*i] = 0.5 * c[i, 0] * twf_old + c[i, 1] * twf_cur

    factor = 2/a1 * (G.L - a2 * sparse.eye(G.N))
    for k in range(2, M):
        twf_new = factor.dot(twf_cur) - twf_old
        for i in range(Nscales):
            r[tmpN + G.N*i] += c[i, k] * twf_new

        twf_old = twf_cur
        twf_cur = twf_new

    return r


def cheby_rect(G, bounds, signal, **kwargs):
    r"""
    Fast filtering using Chebyshev polynomial for a perfect rectangle filter.

    Parameters
    ----------
    G : Graph
    bounds : array_like
        The bounds of the pass-band filter
    signal : array_like
        Signal to filter
    order : int (optional)
        Order of the Chebyshev polynomial (default: 30)

    Returns
    -------
    r : array_like
        Result of the filtering

    """
    if not (isinstance(bounds, (list, np.ndarray)) and len(bounds) == 2):
        raise ValueError('Bounds of wrong shape.')

    bounds = np.array(bounds)

    m = int(kwargs.pop('order', 30) + 1)

    try:
        Nv = np.shape(signal)[1]
        r = np.zeros((G.N, Nv))
    except IndexError:
        r = np.zeros((G.N))

    b1, b2 = np.arccos(2. * bounds / G.lmax - 1.)
    factor = 4./G.lmax * G.L - 2.*sparse.eye(G.N)

    T_old = signal
    T_cur = factor.dot(signal) / 2.
    r = (b1 - b2)/np.pi * signal + 2./np.pi * (np.sin(b1) - np.sin(b2)) * T_cur

    for k in range(2, m):
        T_new = factor.dot(T_cur) - T_old
        r += 2./(k*np.pi) * (np.sin(k*b1) - np.sin(k*b2)) * T_new
        T_old = T_cur
        T_cur = T_new

    return r


def compute_jackson_cheby_coeff(filter_bounds, delta_lambda, m):
    r"""
    To compute the m+1 coefficients of the polynomial approximation of an ideal band-pass between a and b, between a range of values defined by lambda_min and lambda_max.

    Parameters
    ----------
    filter_bounds : list
        [a, b]
    delta_lambda : list
        [lambda_min, lambda_max]
    m : int

    Returns
    -------
    ch : ndarray
    jch : ndarray

    References
    ----------
    :cite:`tremblay2016compressive`

    """
    # Parameters check
    if delta_lambda[0] > filter_bounds[0] or delta_lambda[1] < filter_bounds[1]:
        _logger.error("Bounds of the filter are out of the lambda values")
        raise()
    elif delta_lambda[0] > delta_lambda[1]:
        _logger.error("lambda_min is greater than lambda_max")
        raise()

    # Scaling and translating to standard cheby interval
    a1 = (delta_lambda[1]-delta_lambda[0])/2
    a2 = (delta_lambda[1]+delta_lambda[0])/2

    # Scaling bounds of the band pass according to lrange
    filter_bounds[0] = (filter_bounds[0]-a2)/a1
    filter_bounds[1] = (filter_bounds[1]-a2)/a1

    # First compute cheby coeffs
    ch = np.empty(m+1, dtype=float)
    ch[0] = (2/(np.pi))*(np.arccos(filter_bounds[0])-np.arccos(filter_bounds[1]))
    for i in range(1, len(ch)):
        ch[i] = (2/(np.pi * i)) * \
            (np.sin(i * np.arccos(filter_bounds[0])) - np.sin(i * np.arccos(filter_bounds[1])))

    # Then compute jackson coeffs
    jch = np.empty(m+1, dtype=float)
    alpha = (np.pi/(m+2))
    for i in range(len(jch)):
        jch[i] = (1/np.sin(alpha)) * \
            ((1 - i/(m+2)) * np.sin(alpha) * np.cos(i * alpha) +
             (1/(m+2)) * np.cos(alpha) * np.sin(i * alpha))

    # Combine jackson and cheby coeffs
    jch = ch * jch

    return ch, jch


def lanczos_op(f, s, order=30):
    r"""
    Perform the lanczos approximation of the signal s.

    Parameters
    ----------
    f: Filter
    s : ndarray
        Signal to approximate.
    order : int
        Degree of the lanczos approximation. (default = 30)

    Returns
    -------
    L : ndarray
        lanczos approximation of s

    """
    G = f.G
    Nf = len(f.g)

    # To have the right shape for the output array depending on the signal dim
    try:
        Nv = np.shape(s)[1]
        is2d = True
        c = np.zeros((G.N*Nf, Nv))
    except IndexError:
        Nv = 1
        is2d = False
        c = np.zeros((G.N*Nf))

    tmpN = np.arange(G.N, dtype=int)
    for j in range(Nv):
        if is2d:
            V, H, _ = lanczos(G.L.toarray(), order, s[:, j])
        else:
            V, H, _ = lanczos(G.L.toarray(), order, s)

        Eh, Uh = np.linalg.eig(H)

        Eh[Eh < 0] = 0
        fe = f.evaluate(Eh)
        V = np.dot(V, Uh)

        for i in range(Nf):
            if is2d:
                c[tmpN + i*G.N, j] = np.dot(V, fe[:][i] * np.dot(V.T, s[:, j]))
            else:
                c[tmpN + i*G.N] = np.dot(V, fe[:][i] * np.dot(V.T, s))

    return c


def lanczos(A, order, x):
    r"""
    TODO short description

    Parameters
    ----------
    A: ndarray

    Returns
    -------
    """
    try:
        N, M = np.shape(x)
    except ValueError:
        N = np.shape(x)[0]
        M = 1
        x = x[:, np.newaxis]

    # normalization
    q = np.divide(x, np.kron(np.ones((N, 1)), np.linalg.norm(x, axis=0)))

    # initialization
    hiv = np.arange(0, order*M, order)
    V = np.zeros((N, M*order))
    V[:, hiv] = q

    H = np.zeros((order + 1, M*order))
    r = np.dot(A, q)
    H[0, hiv] = np.sum(q*r, axis=0)
    r -= np.kron(np.ones((N, 1)), H[0, hiv])*q
    H[1, hiv] = np.linalg.norm(r, axis=0)

    orth = np.zeros((order))
    orth[0] = np.linalg.norm(np.dot(V.T, V) - M)

    for k in range(1, order):
        if np.sum(np.abs(H[k, hiv + k - 1])) <= np.spacing(1):
            H = H[:k - 1, _sum_ind(np.arange(k), hiv)]
            V = V[:, _sum_ind(np.arange(k), hiv)]
            orth = orth[:k]

            return V, H, orth

        H[k - 1, hiv + k] = H[k, hiv + k - 1]
        v = q
        q = r/np.tile(H[k - 1, k + hiv], (N, 1))
        V[:, k + hiv] = q

        r = np.dot(A, q)
        r -= np.tile(H[k - 1, k + hiv], (N, 1))*v
        H[k, k + hiv] = np.sum(np.multiply(q, r), axis=0)
        r -= np.tile(H[k, k + hiv], (N, 1))*q

        # The next line has to be checked
        r -= np.dot(V, np.dot(V.T, r))  # full reorthogonalization
        H[k + 1, k + hiv] = np.linalg.norm(r, axis=0)
        orth[k] = np.linalg.norm(np.dot(V.T, V) - M)

    H = H[np.ix_(np.arange(order), np.arange(order))]

    return V, H, orth


def _sum_ind(ind1, ind2):
    ind = np.tile(np.ravel(ind1), (np.size(ind2), 1)).T + np.ravel(ind2)
    return np.ravel(ind)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class Airfoil(Graph):
    r"""Airfoil graph.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Airfoil()
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.5)
    >>> _ = G.plot(edges=True, ax=axes[1])

    """

    def __init__(self, **kwargs):

        data = utils.loadmat('pointclouds/airfoil')
        coords = np.concatenate((data['x'], data['y']), axis=1)

        i_inds = np.reshape(data['i_inds'] - 1, 12289)
        j_inds = np.reshape(data['j_inds'] - 1, 12289)
        A = sparse.coo_matrix((np.ones(12289), (i_inds, j_inds)), shape=(4253, 4253))
        W = (A + A.T) / 2.

        plotting = {"vertex_size": 30,
                    "limits": np.array([-1e-4, 1.01*data['x'].max(),
                                        -1e-4, 1.01*data['y'].max()])}

        super(Airfoil, self).__init__(W, coords=coords, plotting=plotting,
                                      **kwargs)



r"""
Kernel localization
===================

In classical signal processing, a filter can be translated in the vertex
domain. We cannot do that on graphs. Instead, we can
:meth:`~pygsp2.filters.Filter.localize` a filter kernel. Note how on classic
structures (like the ring), the localized kernel is the same everywhere, while
it changes when localized on irregular graphs.
"""

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

fig, axes = plt.subplots(2, 4, figsize=(10, 4))

graphs = [
    pg.graphs.Ring(40),
    pg.graphs.Sensor(64, seed=42),
]

locations = [0, 10, 20]

for graph, axs in zip(graphs, axes):
    graph.compute_fourier_basis()
    g = pg.filters.Heat(graph)
    g.plot(ax=axs[0], title='heat kernel')
    axs[0].set_xlabel(r'eigenvalues $\lambda$')
    axs[0].set_ylabel(r'$g(\lambda) = \exp \left( \frac{{-{}\lambda}}{{\lambda_{{max}}}} \right)$'.format(g.scale[0]))
    maximum = 0
    for loc in locations:
        x = g.localize(loc)
        maximum = np.maximum(maximum, x.max())
    for loc, ax in zip(locations, axs[1:]):
        graph.plot(g.localize(loc), limits=[0, maximum], highlight=loc, ax=ax,
                   title=r'$g(L) \delta_{{{}}}$'.format(loc))
        ax.set_axis_off()

fig.tight_layout()
plt.show()


# -*- coding: utf-8 -*-

import numpy as np

from pygsp2.graphs import NNGraph  # prevent circular import in Python < 3.5


class Sphere(NNGraph):
    r"""Spherical-shaped graph (NN-graph).

    Parameters
    ----------
    radius : float
        Radius of the sphere (default = 1)
    nb_pts : int
        Number of vertices (default = 300)
    nb_dim : int
        Dimension (default = 3)
    sampling : string
        Variance of the distance kernel (default = 'random')
        (Can now only be 'random')
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Sphere(nb_pts=100, seed=42)
    >>> fig = plt.figure()
    >>> ax1 = fig.add_subplot(121)
    >>> ax2 = fig.add_subplot(122, projection='3d')
    >>> _ = ax1.spy(G.W, markersize=1.5)
    >>> _ = _ = G.plot(ax=ax2)

    """

    def __init__(self,
                 radius=1,
                 nb_pts=300,
                 nb_dim=3,
                 sampling='random',
                 seed=None,
                 **kwargs):

        self.radius = radius
        self.nb_pts = nb_pts
        self.nb_dim = nb_dim
        self.sampling = sampling
        self.seed = seed

        if self.sampling == 'random':

            rs = np.random.RandomState(seed)
            pts = rs.normal(0, 1, (self.nb_pts, self.nb_dim))

            for i in range(self.nb_pts):
                pts[i] /= np.linalg.norm(pts[i])

        else:

            raise ValueError('Unknown sampling {}'.format(sampling))

        plotting = {
            'vertex_size': 80,
        }

        super(Sphere, self).__init__(Xin=pts, k=10,
                                     center=False, rescale=False,
                                     plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        attrs = {'radius': '{:.2f}'.format(self.radius),
                 'nb_pts': self.nb_pts,
                 'nb_dim': self.nb_dim,
                 'sampling': self.sampling,
                 'seed': self.seed}
        attrs.update(super(Sphere, self)._get_extra_repr())
        return attrs
